# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from pathlib import Path

from click.testing import CliRunner
import pytest

from swh.graph.cli import graph_cli_group
from swh.graph.example_dataset import DATASET_DIR, RELEASES, REVISIONS

from ..test_cli import read_properties


@pytest.mark.parametrize("workers", [1, 2, 100])
def test_compressgraph(tmpdir, workers):
    tmpdir = Path(tmpdir)

    runner = CliRunner()

    command = [
        "luigi",
        "--base-directory",
        tmpdir / "base_dir",
        "--dataset-name",
        "testdataset",
        "CompressGraph",
        "--batch-size=1000",  # small value, to go fast on the trivial dataset
        "--retry-luigi-delay=1",  # retry quickly
        "--",
        "--CompressGraph-check-flavor",
        "example",
        "--local-scheduler",
        "--CompressGraph-local-export-path",
        DATASET_DIR,
        "--CompressGraph-local-graph-path",
        tmpdir / "compressed_graph",
        "--CompressGraph-rust-executable-dir",
        "./target/debug/",
        f"--workers={workers}",
    ]

    result = runner.invoke(graph_cli_group, command)
    assert result.exit_code == 0, result.stdout

    properties = read_properties(tmpdir / "compressed_graph" / "graph.properties")

    assert int(properties["nodes"]) == 24
    assert int(properties["arcs"]) == 28

    export_meta_path = tmpdir / "compressed_graph/meta/export.json"
    assert (
        export_meta_path.read_bytes() == (DATASET_DIR / "meta/export.json").read_bytes()
    )

    compression_meta_path = tmpdir / "compressed_graph/meta/compression.json"
    compression_meta = json.loads(compression_meta_path.read_text())

    assert compression_meta[0]["conf"]["batch_size"] == 1000
    for step in compression_meta[0]["steps"]:
        assert step["conf"]["batch_size"] == 1000

    assert compression_meta[0]["object_types"] == "cnt,dir,rev,rel,snp,ori"

    with open(
        "swh/graph/example_dataset/compressed/example.property.author_timestamp.bin",
        "rb",
    ) as f:
        timestamps = [
            int.from_bytes(f.read(8), byteorder="big")
            for _ in range(int(properties["nodes"]))
        ]

    # remove non revision/releases
    timestamps = [timestamp for timestamp in timestamps if timestamp != 2**63]
    timestamps.sort()

    expected_timestamps = [
        rel.date.timestamp.seconds for rel in RELEASES if rel.date is not None
    ] + [rev.date.timestamp.seconds for rev in REVISIONS if rev.date is not None]
    expected_timestamps.sort()

    assert timestamps == expected_timestamps


@pytest.mark.parametrize(
    "workers,object_types",
    [
        (1, "ori,snp,rel,rev"),
        (5, "ori,snp,rel,rev"),
        (5, "dir,cnt"),
        (5, "ori,snp"),
        (5, "rel,rev"),
        (5, "rev"),
        (5, "rel"),
    ],
)
def test_compressgraph_partial(tmpdir, workers, object_types):
    tmpdir = Path(tmpdir)

    runner = CliRunner()

    command = [
        "luigi",
        "--base-directory",
        tmpdir / "base_dir",
        "--dataset-name",
        "testdataset",
        "CompressGraph",
        "--batch-size=1000",  # small value, to go fast on the trivial dataset
        "--retry-luigi-delay=1",  # retry quickly
        "--",
        "--CompressGraph-check-flavor",
        "none",
        "--local-scheduler",
        "--CompressGraph-local-export-path",
        DATASET_DIR,
        "--CompressGraph-local-graph-path",
        tmpdir / "compressed_graph",
        f"--CompressGraph-object-types={object_types}",
        "--CompressGraph-rust-executable-dir",
        "./target/debug/",
        f"--workers={workers}",
    ]

    result = runner.invoke(graph_cli_group, command)
    assert result.exit_code == 0, result.stdout

    properties = read_properties(tmpdir / "compressed_graph" / "graph.properties")

    if object_types == "dir,cnt":
        assert int(properties["nodes"]) == 13
        assert int(properties["arcs"]) == 11
    elif object_types == "ori,snp,rel,rev":
        assert int(properties["nodes"]) == 15
        assert int(properties["arcs"]) == 17
    elif object_types == "ori,snp":
        assert int(properties["nodes"]) == 7
        assert int(properties["arcs"]) == 7
        # FIXME: without objects pointed by snapshots, it would be this:
        # assert int(properties["nodes"]) == 4
        # assert int(properties["arcs"]) == 2
    elif object_types == "rel":
        assert int(properties["nodes"]) == 5
        assert int(properties["arcs"]) == 3
        # FIXME: without objects pointed by releases, it would be this:
        # assert int(properties["nodes"]) == 3
        # assert int(properties["arcs"]) == 0
    elif object_types == "rev":
        assert int(properties["nodes"]) == 8
        assert int(properties["arcs"]) == 7
        # FIXME: without objects pointed by revisions, it would be this:
        # assert int(properties["nodes"]) == 4
        # assert int(properties["arcs"]) == 3
    elif object_types == "rel,rev":
        assert int(properties["nodes"]) == 11
        assert int(properties["arcs"]) == 10
        # FIXME: without objects pointed by revisions, it would be this:
        # assert int(properties["nodes"]) == 7
        # assert int(properties["arcs"]) == 7
    else:
        assert False, f"Unexpected object types {object_types}"

    export_meta_path = tmpdir / "compressed_graph/meta/export.json"
    assert (
        export_meta_path.read_bytes() == (DATASET_DIR / "meta/export.json").read_bytes()
    )

    compression_meta_path = tmpdir / "compressed_graph/meta/compression.json"
    compression_meta = json.loads(compression_meta_path.read_text())

    assert compression_meta[0]["conf"]["batch_size"] == 1000
    for step in compression_meta[0]["steps"]:
        assert step["conf"]["batch_size"] == 1000

    assert compression_meta[0]["object_types"] == object_types



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Spawns a gRPC server on a newly compressed graph, and checks responses
are as expected."""

from base64 import b64decode
import logging
from pathlib import Path
import socket
import time
from typing import Literal

import attr
from google.protobuf.field_mask_pb2 import FieldMask
import grpc

from swh.graph.config import check_config_compress
import swh.graph.grpc.swhgraph_pb2 as swhgraph
import swh.graph.grpc.swhgraph_pb2_grpc as swhgraph_grpc
from swh.graph.grpc_server import spawn_rust_grpc_server, stop_grpc_server
from swh.graph.shell import Rust, Sink
from swh.model.swhids import CoreSWHID, QualifiedSWHID

e2e_logger = logging.getLogger(__name__)


def run_e2e_check(
    graph_name: str,
    in_dir: str | None,
    out_dir: str | None,
    sensitive_in_dir: str | None,
    sensitive_out_dir: str | None,
    check_flavor: Literal["full", "history_hosting", "staging", "example", "none"],
    profile: Literal["release", "debug"] = "release",
    logger: logging.Logger | None = None,
):
    """Empirically check the graph compression correctness.

    Check for specific SWHIDs in the compressed graph and do simple traversal
    requests to ensure the compression went well.

    Args:
        graph_name: Graph base name, relative to in_dir.
        in_dir: Input directory, where the uncompressed graph can be found.
        out_dir: Output directory, where the compressed graph can be found.
        sensitive_in_dir: Sensitive input directory, where the uncompressed
            sensitive graph can be found.
        sensitive_out_dir: Sensitive output directory, where the compressed
            sensitive graph can be found.
        check_flavor: Which flavor of checks to run.

            - "full": for a dataset made from the entire archive
            - "history_hosting": for a history and hosting dataset
            - "staging": for a dataset made from staging
            - "example": for the example dataset shipped with `swh-graph`
            - "none": for preventing to run the checks altogether

        profile: Which Rust executables to use.

    Raises:
        Exception: GRPC server unexpectedly stopped.
    """
    if logger is None:
        logger = e2e_logger

    conf = check_config_compress(
        {"profile": profile},
        graph_name,
        in_dir,
        out_dir,
        sensitive_in_dir,
        sensitive_out_dir,
        check_flavor,
    )

    graph_name = conf["graph_name"]
    in_dir = conf["in_dir"]
    out_dir = conf["out_dir"]
    check_flavor = conf["check_flavor"]

    if check_flavor == "none":
        logger.info("End-to-end checks skipped.")
        return

    if "graph_path" not in conf:
        conf["graph_path"] = f"{out_dir}/{graph_name}"

    server, port = spawn_rust_grpc_server(**conf, path=conf["graph_path"])

    # wait for the server to accept connections
    while True:
        try:
            socket.create_connection(("localhost", port))
        except ConnectionRefusedError:
            time.sleep(0.1)
            server.poll()
            if server.returncode is not None:
                raise Exception("GRPC server unexpectedly stopped.")
        else:
            break

    # This dictionary is a copy of a few items retrieved from the graph.
    # It includes a few well known projects, such as `parmap`, `apt`, or `vim`.
    # They should not change in the future, as all traversals start from fixed snapshots.
    # The keys of this dictionary are the origins of the aforementioned projects.
    # For each origin the there is a dictionary where the keys are the SWHIDs of snapshots
    # of said origin.
    # For each snapshot, there is a list of SWHIDs that should be found in this snapshot.
    # The revisions all have the same author, but not necessarily the same committer.
    # The directory is the root directory of the project at the latest revision found in its
    # corresponding snapshot. The fact that it is the root directory is not checked here.
    # The content is the README of the project at the latest revision found in its
    # corresponding snapshot. The fact that it's a README is not checked here.
    projects: dict[str, dict[str, list[str]]] = {
        "https://github.com/rdicosmo/parmap": {
            "swh:1:snp:01b2cc89f4c423f1bda4757edd86ae4013b919b0": [
                "swh:1:rev:963608763589e03de38e744d359884d491e65460",
                "swh:1:rev:db44dc9cf7a6af7b56d8ebda8c75be3375c89282",
                "swh:1:rev:1846939e731b0f5552e2cdf3b0667042cb3d5b30",
                "swh:1:rev:ba9a61336ca4ad78012566e506eebf343ef1ac1f",
                "swh:1:rev:36089bc12f3560faffeeb66d3f38594978561c70",
                "swh:1:dir:2dc0f462d191524530f5612d2935851505af41dd",
                "swh:1:cnt:43243e2ae91a64e252170cd922718e8c2af323b6",
            ],
        },
        "https://github.com/pallets/flask": {
            "swh:1:snp:e8a88d27c5f999868004a3186dc753f6b64f52fa": [
                "swh:1:rev:4fe0aebab79a092615f5f86a24b91bac07fb2ef2",
                "swh:1:rev:e3535f99717bdf3882c0b48bff3e3eb244454e34",
                "swh:1:rev:f567ab90681235f38bdeaeda004d89e85b8627d9",
                "swh:1:rev:7621b3d96ac2d53846c4df211bc05c5e13420028",
                "swh:1:rev:e165f3aef4dbede5d7e3f71c336c89a268753ae9",
                "swh:1:dir:a098f5ea0c6dc036256d936d6da119d28dcf8e86",
                "swh:1:cnt:df4d41cbd3779eb7f865f3b357546acdbc59fa00",
            ],
        },
        "https://github.com/home-assistant/core": {
            "swh:1:snp:4a4962aaf6939bcc2a50f5d3f8fef1fb8ae482db": [
                "swh:1:rev:fa8284d360e64037aa06b1081bdb0f655d9d7be8",
                "swh:1:rev:9b3f92e265937a18858fef2f7b26a6e3fed8998b",
                "swh:1:rev:8b46c8bf206e860a36351de84ec505e65adc7bfa",
                "swh:1:rev:77d83bffee3c767b0014a3c4296e18002efc1549",
                "swh:1:rev:6df77ef94ba90ba6b3c2c9530be0ba4f61c43558",
                "swh:1:dir:b711bde9af241d46fdf7fc63924eecef717a16ec",
                "swh:1:cnt:85c632f7eb1d1861911c735077699c29840ff728",
            ]
        },
        "https://github.com/neovim/neovim": {
            "swh:1:snp:296e07d432c81a8c34d4418f10848af0ed8cb4b8": [
                "swh:1:rev:d82168e41c21a3a107e52c139abc90dbe41f2010",
                "swh:1:rev:046340303076c3680cf767f01c6748d9dd5c8fa5",
                "swh:1:rev:7db9992445a69df16eb12d349d49ab8968e9a505",
                "swh:1:rev:64c2c7c40da4bafb6f74076b7ffbffb262246c7a",
                "swh:1:rev:788bc12a6f4c5a4627cbc75a2f539bfc622384a2",
                "swh:1:dir:234c61bf6e96643c59a33586039c82b492ca86af",
                "swh:1:cnt:8feb7918886d54844bf9234a314f28675b9a76fc",
            ]
        },
        "https://gitlab.softwareheritage.org/swh/devel/swh-graph": {
            "swh:1:snp:d34d87373bb367ba310002693cb7c4c139c3b882": [
                "swh:1:rev:985dcf705e03fde55285ca8aaff2488f43e9a55f",
                "swh:1:rev:af36dc6f7a6d69fbcceccf7f6ef8cea626e06d55",
                "swh:1:rev:b777c9d0d7405cd25f0cf3f60bb21ec38d6eeb58",
                "swh:1:rev:a894a2b14e892fe3f3c126a0ceb4299020f7877d",
                "swh:1:rev:5109f7788c4a59443dd5beeef010dded3f3a1b70",
                "swh:1:dir:1618172adfdc2baef53ff0c27ebbd33899c455b0",
                "swh:1:cnt:d7abf98f65283673a669eeba114f545934153f5d",
            ],
        },
    }

    # This is a dictionary holding the full names of the authors of the revisions included
    # in the `projects` dictionary.
    fullnames: dict[str, str] = {
        "https://github.com/rdicosmo/parmap": (
            "Um9iZXJ0byBEaSBDb3NtbyA8cm9iZXJ0b0BkaWNvc21vLm9yZz4="
        ),
        "https://github.com/pallets/flask": (
            "ZGVwZW5kYWJvdFtib3RdIDw0OTY5OTMzMytkZXBlbmRhYm90W2JvdF1AdXNlcnMubm9yZXBseS5naXRodWIuY29tPg=="
        ),
        "https://github.com/home-assistant/core": (
            "ZGVwZW5kYWJvdFtib3RdIDw0OTY5OTMzMytkZXBlbmRhYm90W2JvdF1AdXNlcnMubm9yZXBseS5naXRodWIuY29tPg=="
        ),
        "https://github.com/neovim/neovim": (
            "ZGVwZW5kYWJvdFtib3RdIDw0OTY5OTMzMytkZXBlbmRhYm90W2JvdF1AdXNlcnMubm9yZXBseS5naXRodWIuY29tPg=="
        ),
        "https://gitlab.softwareheritage.org/swh/devel/swh-graph": (
            "VmFsZW50aW4gTG9yZW50eiA8dmxvcmVudHpAc29mdHdhcmVoZXJpdGFnZS5vcmc+"
        ),
    }

    errors: list[str | int | QualifiedSWHID] = []

    # This is a compressed graph of only the “history and hosting” layer (origins,
    # snapshots, releases, revisions) and the root directory (or rarely content) of every
    # revision/release; but most directories and contents are excluded.
    if check_flavor == "history_hosting":
        for project in projects.values():
            for swhids in project.values():
                for swhid in swhids:
                    if swhid.startswith("swh:1:dir:") or swhid.startswith("swh:1:cnt:"):
                        swhids.remove(swhid)
    # Before October 2025, only `parmap` and `apt` were in staging, so other origins need to
    # be removed for the checks to succeed.
    elif check_flavor == "staging":
        for origin in [
            "https://github.com/pallets/flask",
            "https://github.com/home-assistant/core",
            "https://github.com/neovim/neovim",
            "https://gitlab.softwareheritage.org/swh/devel/swh-graph",
        ]:
            projects.pop(origin)
            fullnames.pop(origin)
    # This is the example graph (`swh/graph/example_dataset/`). It is
    # mostly used for CI and testing purposes.
    elif check_flavor == "example":
        projects = {
            "https://example.com/swh/graph": {
                "swh:1:snp:0000000000000000000000000000000000000022": [
                    "swh:1:rev:0000000000000000000000000000000000000009",
                    "swh:1:dir:0000000000000000000000000000000000000006",
                    "swh:1:cnt:0000000000000000000000000000000000000004",
                ],
            }
        }

    # All authors of listed revisions should be the same, but we don't know their IDs
    # so we can only check whether they all have the same ID or not.
    authors: dict[str, int | None] = {}

    try:
        with grpc.insecure_channel(f"localhost:{port}") as channel:
            stub = swhgraph_grpc.TraversalServiceStub(channel)
            for origin, project in projects.items():
                for snp_swhid, swhids in project.items():
                    response = stub.Traverse(
                        swhgraph.TraversalRequest(
                            src=[snp_swhid],
                            mask=FieldMask(paths=["swhid", "rev.author"]),
                        )
                    )
                    for elt in response:
                        if elt.swhid in swhids:
                            swhids.remove(elt.swhid)
                            if elt.swhid.startswith("swh:1:rev:"):
                                # Add the author ID of the revision's author as the value
                                # of `authors[origin]` if there isn't one already, otherwise
                                # checks they are the same.
                                if (
                                    authors.setdefault(origin, elt.rev.author)
                                    != elt.rev.author
                                ):
                                    full_swhid = attr.evolve(
                                        QualifiedSWHID.from_string(elt.swhid),
                                        origin=origin,
                                        visit=CoreSWHID.from_string(snp_swhid),
                                    )
                                    logger.error(f"Author ID for {full_swhid} is wrong")
                                    errors.append(full_swhid)
    finally:
        stop_grpc_server(server)

    # Check if the author IDs previously checked match their full names. This is triggered
    # only when the sensitive files containing said full names are present on disk.
    if (
        Path(f"{sensitive_out_dir}/{graph_name}.persons").exists()
        and Path(f"{sensitive_out_dir}/{graph_name}.persons.ef").exists()
    ):
        for origin, author in authors.items():
            if author is None:
                return
            fullname = (
                (
                    Rust(
                        "swh-graph-person-id-to-name",
                        f"{author}",
                        f"{sensitive_out_dir}/{graph_name}",
                    )
                    > Sink()
                )
                .run()
                .stdout
            )
            if fullname.removesuffix(b"\n") != b64decode(fullnames[origin]):
                logger.error(
                    f"Full name for author ID {author} for project {origin} is incorrect: "
                    f"expected {b64decode(fullnames[origin]).decode()}, "
                    f"got {fullname.decode(errors='replace')}"
                )
                errors.append(author)
    else:
        logger.warn(
            f"Could not find {sensitive_out_dir}/{graph_name}.persons or "
            f"{sensitive_out_dir}/{graph_name}.persons.ef.\n"
            "End-to-end checks for full names skipped"
        )

    for origin, project in projects.items():
        for snp_swhid, swhids in project.items():
            for swhid in swhids:
                full_swhid = attr.evolve(
                    QualifiedSWHID.from_string(swhid),
                    origin=origin,
                    visit=CoreSWHID.from_string(snp_swhid),
                )
                logger.error(f"{full_swhid} has not been found during the traversal")
                errors.append(full_swhid)

    if errors:
        error_msg = "\n".join([str(error) for error in errors])
        raise Exception(
            "End-to-end checks for compression failed. "
            "The following errors have been detected:\n"
            f"{error_msg}\n"
            "See above logs for more details"
        )
    logger.info("Compression seems to have gone well")



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
from pathlib import Path
from typing import Dict, List

import luigi
import requests

from swh.export.luigi import (
    AthenaDatabaseTarget,
    CreateAthena,
    ObjectType,
    S3PathParameter,
    merge_lists,
)

from .compressed_graph import _tables_for_object_types


class SelectTopGithubOrigins(luigi.Task):
    """Writes a list of origins selected from popular Github repositories"""

    local_export_path = luigi.PathParameter()
    num_origins = luigi.IntParameter(default=1)
    query = luigi.Parameter(
        default="language:python",
        description="Search query to use to filter Github repositories",
    )

    def output(self) -> luigi.LocalTarget:
        """Text file with a list of origin URLs"""
        return luigi.LocalTarget(self.local_export_path / "origins.txt")

    def run(self) -> None:
        """Sends a query to the Github API to get a list of origins"""
        import math

        urls: List[str] = []
        for i in range(0, math.ceil(self.num_origins / 100)):
            resp = requests.get(
                "https://api.github.com/search/repositories",
                params=dict(  # type: ignore[arg-type]
                    page=i,
                    s="starts",
                    order="desc",
                    q=self.query,
                    per_page=min(100, self.num_origins - len(urls)),
                ),
            )
            urls.extend(repo["html_url"] for repo in resp.json()["items"])

        with self.output().open("wt") as fd:
            for url in urls:
                fd.write(f"{url}\n")


class SubdatasetOriginsFromFile(luigi.Task):
    """Reads a list of origins from a local file, computed externally to Luigi."""

    local_export_path = luigi.PathParameter()
    path = luigi.Parameter(
        default="",
        description="What file to read origins from. "
        "Defaults to local_export_path / origins.txt",
    )

    def output(self) -> luigi.LocalTarget:
        """Text file with a list of origin URLs"""
        return luigi.LocalTarget(self.path or self.local_export_path / "origins.txt")

    def run(self) -> None:
        """Does nothing"""
        pass


class ListSwhidsForSubdataset(luigi.Task):
    """Lists all SWHIDs reachable from a set of origins"""

    select_task = luigi.ChoiceParameter(
        choices=["SelectTopGithubOrigins", "SubdatasetOriginsFromFile"],
        default="SelectTopGithubOrigins",
        description="Which algorithm to use to generate the list of origins",
    )
    local_export_path = luigi.PathParameter()
    grpc_api = luigi.Parameter()

    def requires(self) -> luigi.Task:
        """Returns an instance of ``self.select_task``"""
        return globals()[self.select_task](local_export_path=self.local_export_path)

    def output(self) -> luigi.LocalTarget:
        """Text file with a list of SWHIDs"""
        return luigi.LocalTarget(self.local_export_path / "swhids.txt")

    def run(self) -> None:
        """Builds the list"""
        import hashlib

        from google.protobuf.field_mask_pb2 import FieldMask
        import grpc
        import tqdm

        import swh.graph.grpc.swhgraph_pb2 as swhgraph
        import swh.graph.grpc.swhgraph_pb2_grpc as swhgraph_grpc

        with self.input().open("r") as fd:
            origin_urls = [line.strip() for line in fd]

        origin_swhids = [
            f"swh:1:ori:{hashlib.sha1(url.encode()).hexdigest()}" for url in origin_urls
        ]

        with grpc.insecure_channel(self.grpc_api) as channel:
            stub = swhgraph_grpc.TraversalServiceStub(channel)

            known_origin_swhids = []
            for origin_swhid in origin_swhids:
                try:
                    stub.GetNode(swhgraph.GetNodeRequest(swhid=origin_swhid))
                except grpc.RpcError as e:
                    if e.code() != grpc.StatusCode.NOT_FOUND:
                        raise
                else:
                    known_origin_swhids.append(origin_swhid)

            unknown_origins = set(origin_swhids) - set(known_origin_swhids)
            print(f"Filtered out {len(unknown_origins)} unknown origins")

            request = swhgraph.TraversalRequest(
                src=known_origin_swhids,
                mask=FieldMask(paths=["swhid"]),
            )

            with self.output().open("wt") as fd:
                for item in tqdm.tqdm(stub.Traverse(request)):
                    fd.write(f"{item.swhid}\n")


class CreateSubdatasetOnAthena(luigi.Task):
    """Generates an ORC export from an existing ORC export, filtering out SWHIDs
    not in the given list."""

    local_export_path: Path = luigi.PathParameter()
    s3_parent_export_path: str = S3PathParameter(  # type: ignore[assignment]
        description="s3:// URL to the existing complete export",
    )
    s3_export_path: str = S3PathParameter(  # type: ignore[assignment]
        description="s3:// URL to the export to produce",
    )
    s3_athena_output_location = S3PathParameter()
    athena_db_name = luigi.Parameter()
    athena_parent_db_name = luigi.Parameter()
    object_types = luigi.EnumListParameter(
        enum=ObjectType, default=list(ObjectType), batch_method=merge_lists
    )

    def requires(self) -> Dict[str, luigi.Task]:
        """Returns an instance of :class:`ListSwhidsForSubdataset` and one of
        :class:`CreateAthena`"""
        return {
            "swhids": ListSwhidsForSubdataset(local_export_path=self.local_export_path),
            "export": CreateAthena(
                athena_db_name=self.athena_parent_db_name,
                s3_export_path=self.s3_parent_export_path,
            ),
        }

    def output(self) -> Dict[str, luigi.Target]:
        """Returns the S3 location and Athena database for the subdataset"""
        from swh.export.athena import TABLES

        return {
            "orc": self._meta(),
            "athena": AthenaDatabaseTarget(self.athena_db_name, set(TABLES)),
        }

    def _meta(self):
        import luigi.contrib.s3

        return luigi.contrib.s3.S3Target(f"{self.s3_export_path}/meta/export.json")

    def run(self) -> None:
        """Runs a query on Athena, producing files on S3"""
        import datetime
        from importlib.metadata import version
        import json
        import socket

        from swh.export.athena import generate_subdataset

        start_date = datetime.datetime.now(tz=datetime.timezone.utc)
        generate_subdataset(
            self.athena_parent_db_name,
            self.athena_db_name,
            self.s3_export_path.rstrip("/") + "/orc",
            self.input()["swhids"].path,
            self.s3_export_path.rstrip("/") + "/queries",
        )
        end_date = datetime.datetime.now(tz=datetime.timezone.utc)

        with luigi.contrib.s3.S3Target(
            f"{self.s3_parent_export_path}/meta/export.json"
        ).open("r") as fd:
            parent_meta = json.load(fd)

        meta = {
            "flavor": "subdataset",
            "export_start": start_date.isoformat(),
            "export_end": end_date.isoformat(),
            "object_types": _tables_for_object_types(self.object_types),
            "parent": parent_meta,
            "hostname": socket.getfqdn(),
            "tool": {
                "name": "swh.export",
                "version": version("swh.export"),
            },
        }

        with self._meta().open("w") as fd:
            json.dump(meta, fd, indent=4)



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import subprocess
import threading

import pytest
import pyzstd

from swh.graph.shell import AtomicFileSink, Command, CommandException, Pipe, Sink, wc

# fmt: off


def test_basic_stdout():
    assert (Command.echo("foo") > Sink()).run().stdout == b"foo\n"


def test_basic_file(tmp_path):
    path = tmp_path / "file.txt"
    (
        Command.echo("foo")
        > AtomicFileSink(path)
    ).run()
    assert path.read_bytes() == b"foo\n"


def test_pipe_stdout():
    res = (
        Command.echo("foo")
        | Command.zstdmt()
        > Sink()
    ).run().stdout
    assert pyzstd.decompress(res) == b"foo\n", res


def test_large_sink():
    """Checks Sink() does not block"""
    res = None

    def f():
        nonlocal res
        res = (
            Command.yes()
            | Command.head("-n", "1000000")
            > Sink()
        ).run().stdout
        assert res == b"y\n" * 1000000, res

    thread = threading.Thread(target=f)
    thread.start()
    thread.join(10)  # 0.1s should be enough, but let's avoid flaky tests
    assert not thread.is_alive(), "blocked or took too long"


def test_pipe_file(tmp_path):
    path = tmp_path / "file.txt"
    (
        Command.echo("foo")
        | Command.zstdmt()
        > AtomicFileSink(path)
    ).run()
    res = path.read_bytes()
    print(res)  # in case pyzstd.decompress() fails
    assert pyzstd.decompress(res) == b"foo\n", res


def test_error(tmp_path):
    path = tmp_path / "bar.txt"

    with pytest.raises(CommandException):
        (
            Command.cat(tmp_path / "foo.txt")
            > AtomicFileSink(path)
        ).run()

    assert not path.exists()
    assert not (path / "bar.txt.tmp").exists()


def test_concat_command(tmp_path):
    path = tmp_path / "foo.txt"
    path.write_bytes(b"foo\n")
    assert (
        Command.echo("bar")
        | Command.cat(path, "-", Command.echo("baz"))
        > Sink()
    ).run().stdout == b"foo\nbar\nbaz\n"


def test_wc(tmp_path):
    path = tmp_path / "foo.txt"
    path.write_bytes(b"foo\nbar\nbaz\n")
    assert wc(Command.cat(path), "-l") == 3


def test_pv_wc(tmp_path):
    path = tmp_path / "foo.txt"
    path.write_bytes(b"foo\nbar\nbaz\n")
    assert (
        Command.cat(path)
        | Command.pv(
            "--line-mode",
            "--size",
            str(wc(Command.cat(path), "-l"))
        )
        > Sink()
    ).run().stdout == b"foo\nbar\nbaz\n"


def test_concat_pipe(tmp_path):
    path = tmp_path / "foo.txt"
    path.write_bytes(pyzstd.compress(b"foo\n"))
    res = (
        Command.echo("bar")
        | Command.zstdmt()
        | Command.cat(path, "-", Command.echo("baz") | Command.zstdmt())
        > Sink()
    ).run().stdout
    print(res)  # in case pyzstd.decompress() fails
    assert pyzstd.decompress(res) == b"foo\nbar\nbaz\n"


def test_associativity(tmp_path):
    assert (
        Command.echo("foo")
        | Command.cat()
        | Command.cat()
        > Sink()
    ).run().stdout == b"foo\n"

    assert (
        (
            Command.echo("foo")
            | Command.cat()
        )
        | Command.cat()
        > Sink()
    ).run().stdout == b"foo\n"

    assert (
        Command.echo("foo")
        | (
            Command.cat()
            | Command.cat()
        )
        > Sink()
    ).run().stdout == b"foo\n"


def test_file_atomic(tmp_path):
    path = tmp_path / "file.txt"

    for content in ("foo", "bar"):
        (
            Command.echo(content)
            | Command.cat("-", Command.true())
            > AtomicFileSink(path)
        ).run()
        assert path.read_bytes() == f"{content}\n".encode()

    with pytest.raises(CommandException):
        (
            Command.echo("qux")
            | Command.cat("-", Command.false())
            > AtomicFileSink(path)
        ).run()
    assert path.read_bytes() == b"bar\n"  # wasn't overwritten because the pipe failed

    assert not (tmp_path / "file.txt.tmp").exists(), "temporary file was not cleaned up"


def test_atomic_file_sink_run_with_stderr_fd(tmp_path):
    path = tmp_path / "file.txt"
    try:
        try:
            (r, w) = os.pipe()
            (
                Command.bash("-c", "echo foo; >&2 echo bar")
                | Command.zstdmt()
                > AtomicFileSink(path)
            ).run(stderr=w)
        finally:
            os.close(w)
        res = path.read_bytes()
        assert pyzstd.decompress(res) == b"foo\n", res
    except Exception:
        os.close(r)
        raise
    with os.fdopen(r) as read_fd:
        assert read_fd.read() == "bar\n"


def test_atomic_file_sink__run_with_stderr_pipe(tmp_path):
    path = tmp_path / "file.txt"
    cmd = (
        Command.bash("-c", "echo foo; >&2 echo bar")
        | Command.zstdmt()
        > AtomicFileSink(path)
    )._run(stdin=None, stdout=None, stderr=subprocess.PIPE)
    cmd.wait()
    res = path.read_bytes()
    assert pyzstd.decompress(res) == b"foo\n", res
    assert cmd.stderr().read() == b"bar\n"


def test_atomic_file_sink__run_with_stderr_is_stdout(tmp_path):
    path = tmp_path / "file.txt"
    cmd = (
        Command.bash("-c", "echo foo; >&2 echo bar")
        | Command.cat()
        > AtomicFileSink(path)
    )._run(stdin=None, stdout=None, stderr=subprocess.STDOUT)
    cmd.wait()
    res = path.read_bytes()
    assert res == b"foo\n", res
    assert cmd.stdout().read() == b"bar\n"


def test_command__run_with_pipe_and_stdout():
    cmd = (
        Command.bash("-c", "echo foo; >&2 echo bar")
    )._run(stdin=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    cmd.wait()
    assert sorted(cmd.stdout().read().splitlines()) == [b"bar", b"foo"]


def test_command__run_with_same_fd():
    try:
        try:
            (r, w) = os.pipe()
            cmd = (
                Command.bash("-c", "echo foo; >&2 echo bar")
            )._run(stdin=None, stdout=w, stderr=w)
        finally:
            os.close(w)
        cmd.wait()
    except Exception:
        os.close(r)
        raise
    with os.fdopen(r) as read_fd:
        assert sorted(read_fd.read().splitlines()) == ["bar", "foo"]


def test_command__run_with_different_fd():
    try:
        try:
            (r1, w1) = os.pipe()
            (r2, w2) = os.pipe()
            cmd = (
                Command.bash("-c", "echo foo; >&2 echo bar")
            )._run(stdin=None, stdout=w1, stderr=w2)
        finally:
            os.close(w1)
            os.close(w2)
        cmd.wait()
    except Exception:
        os.close(r1)
        raise
    except Exception:
        os.close(r2)
        raise
    with os.fdopen(r1) as read_fd_1:
        assert sorted(read_fd_1.read().splitlines()) == ["foo"]
    with os.fdopen(r2) as read_fd_2:
        assert sorted(read_fd_2.read().splitlines()) == ["bar"]


def test_pipe__run_with_pipe_and_stdout():
    cmd = (
        Pipe([
            Command.bash("-c", "echo foo; >&2 echo bar"),
            Command.cat(),
        ])._run(stdin=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    )
    cmd.wait()
    assert sorted(cmd.stdout().read().splitlines()) == [b"bar", b"foo"]


def test_pipe__run_with_same_fd():
    try:
        try:
            (r, w) = os.pipe()
            cmd = (
                Pipe([
                    Command.bash("-c", "echo foo; >&2 echo bar"),
                    Command.cat(),
                ])._run(stdin=None, stdout=w, stderr=w)
            )
        finally:
            os.close(w)
        cmd.wait()
    except Exception:
        os.close(r)
        raise
    with os.fdopen(r) as read_fd:
        assert sorted(read_fd.read().splitlines()) == ["bar", "foo"]


def test_pipe__run_with_different_fd():
    try:
        try:
            (r1, w1) = os.pipe()
            (r2, w2) = os.pipe()
            cmd = (
                Pipe([
                    Command.bash("-c", "echo foo; >&2 echo bar"),
                    Command.cat(),
                ])._run(stdin=None, stdout=w1, stderr=w2)
            )
        finally:
            os.close(w1)
            os.close(w2)
        cmd.wait()
    except Exception:
        os.close(r1)
        raise
    except Exception:
        os.close(r2)
        raise
    with os.fdopen(r1) as read_fd_1:
        assert sorted(read_fd_1.read().splitlines()) == ["foo"]
    with os.fdopen(r2) as read_fd_2:
        assert sorted(read_fd_2.read().splitlines()) == ["bar"]


def test_sink_run_with_stdout():
    with pytest.raises(NotImplementedError):
        (
            Command.bash("-c", "echo foo; >&2 echo bar")
            > Sink()
        ).run(stderr=subprocess.STDOUT)


def test_sink__run_with_stderr_fd():
    try:
        try:
            (r, w) = os.pipe()
            res = (
                Command.bash("-c", "echo foo; >&2 echo bar")
                > Sink()
            ).run(stderr=w)
        finally:
            os.close(w)
        assert res.stdout == b"foo\n"
    except Exception:
        os.close(r)
        raise
    with os.fdopen(r) as read_fd:
        assert read_fd.read() == "bar\n"


def test_sink__run_with_pipes():
    with pytest.raises(NotImplementedError):
        (
            Command.bash("-c", "echo foo; >&2 echo bar")
            > Sink()
        ).source_pipe._run(
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )


def test_process_substitution_no_stderr():
    res = (
        Command.cat(
            Command.bash("-c", "echo foo1; >&2 echo bar1"),
            Command.bash("-c", "echo foo2; >&2 echo bar2"),
        )
        > Sink()
    ).run()
    assert res.stdout == b"foo1\nfoo2\n"


def test_process_substitution_with_stderr_fd():
    try:
        try:
            (r, w) = os.pipe()
            res = (
                Command.cat(
                    Command.bash("-c", "echo foo1; >&2 echo bar1"),
                    Command.bash("-c", "echo foo2; >&2 echo bar2"),
                )
                > Sink()
            ).run(stderr=w)
        finally:
            os.close(w)
        assert res.stdout == b"foo1\nfoo2\n"
    except Exception:
        os.close(r)
        raise
    with os.fdopen(r, "rb") as read_fd:
        assert sorted(read_fd.read().splitlines()) == [b"bar1", b"bar2"]


def test_process_substitution_with_stderr_to_stdout():
    with pytest.raises(NotImplementedError):
        (
            Command.cat(
                Command.bash("-c", "echo foo1; >&2 echo bar1"),
                Command.bash("-c", "echo foo2; >&2 echo bar2"),
            )
            > Sink()
        ).run(stderr=subprocess.STDOUT)



#  Copyright (c) 2022 The Software Heritage developers
#  See the AUTHORS file at the top-level directory of this distribution
#  License: GNU General Public License version 3, or any later version
#  See top-level LICENSE file for more information


import warnings

from .http_client import *  # noqa

warnings.warn(
    "the swh.graph.client module is deprecated, use swh.graph.http_client instead",
    DeprecationWarning,
    stacklevel=2,
)



# Copyright (C) 2019-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Script to run the end to end compression checks without having to
compress an entire graph.
"""

from swh.graph.e2e_check import run_e2e_check

run_e2e_check(
    graph_name="example",
    in_dir="swh/graph/example_dataset/orc/",
    out_dir="swh/graph/example_dataset/compressed/",
    sensitive_in_dir="swh/graph/example_dataset-sensitive/orc/",
    sensitive_out_dir="swh/graph/example_dataset-sensitive/compressed/",
    check_flavor="example",
    profile="debug",
)



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from pathlib import Path
import shlex
import sys
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import click

from swh.core.cli import CONTEXT_SETTINGS, AliasedGroup
from swh.core.cli import swh as swh_cli_group

if TYPE_CHECKING:
    from swh.graph.webgraph import CompressionStep  # noqa

logger = logging.getLogger(__name__)


class StepOption(click.ParamType):
    """click type for specifying a compression step on the CLI

    parse either individual steps, specified as step names or integers, or step
    ranges

    """

    name = "compression step"

    def convert(self, value, param, ctx):  # type: (...) -> Set[CompressionStep]
        from swh.graph.webgraph import COMP_SEQ, CompressionStep  # noqa

        steps: Set[CompressionStep] = set()

        specs = value.split(",")
        for spec in specs:
            if "-" in spec:  # step range
                (raw_l, raw_r) = spec.split("-", maxsplit=1)
                if raw_l == "":  # no left endpoint
                    raw_l = COMP_SEQ[0].name
                if raw_r == "":  # no right endpoint
                    raw_r = COMP_SEQ[-1].name
                l_step = self.convert(raw_l, param, ctx)
                r_step = self.convert(raw_r, param, ctx)
                if len(l_step) != 1 or len(r_step) != 1:
                    self.fail(f"invalid step specification: {value}, " f"see --help")
                l_idx = l_step.pop()
                r_idx = r_step.pop()
                steps = steps.union(
                    set(CompressionStep(i) for i in range(l_idx.value, r_idx.value + 1))
                )
            else:  # singleton step
                try:
                    steps.add(CompressionStep(int(spec)))  # integer step
                except ValueError:
                    try:
                        steps.add(CompressionStep[spec.upper()])  # step name
                    except KeyError:
                        self.fail(
                            f"invalid step specification: {value}, " f"see --help"
                        )

        return steps


class PathlibPath(click.Path):
    """A Click path argument that returns a pathlib Path, not a string"""

    def convert(self, value, param, ctx):
        return Path(super().convert(value, param, ctx))


DEFAULT_CONFIG: Dict[str, Tuple[str, Any]] = {
    "graph": ("dict", {"cls": "local", "grpc_server": {}})
}


@swh_cli_group.group(name="graph", context_settings=CONTEXT_SETTINGS, cls=AliasedGroup)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="YAML configuration file",
)
@click.option(
    "--profile",
    type=str,
    help="Which Rust profile to use executables from, usually 'release' "
    "(the default) or 'debug'.",
)
@click.pass_context
def graph_cli_group(ctx, config_file, profile):
    """Software Heritage graph tools."""
    from swh.core import config

    ctx.ensure_object(dict)
    conf = config.read(config_file, DEFAULT_CONFIG)
    if "graph" not in conf:
        raise ValueError(
            'no "graph" stanza found in configuration file %s' % config_file
        )
    ctx.obj["config"] = conf

    if profile is not None:
        conf["profile"] = profile


@graph_cli_group.command(name="rpc-serve")
@click.option(
    "--host",
    "-h",
    default="0.0.0.0",
    metavar="IP",
    show_default=True,
    help="host IP address to bind the server on",
)
@click.option(
    "--port",
    "-p",
    default=5009,
    type=click.INT,
    metavar="PORT",
    show_default=True,
    help="port to bind the server on",
)
@click.option(
    "--graph",
    "-g",
    "graph_path",
    default=None,
    metavar="GRAPH",
    help="compressed graph basename",
)
@click.pass_context
def serve(ctx, host, port, graph_path):
    """Run the compressed graph HTTP RPC service.

    The documentation of the HTTP RPC API is available on
    https://docs.softwareheritage.org/devel/swh-graph/api.html
    """
    import aiohttp.web

    from swh.graph.http_rpc_server import make_app

    config = ctx.obj["config"]
    config_graph = config["graph"]
    cls = config_graph["cls"]
    if cls in ("local", "local_rust"):
        if graph_path is None:
            raise ValueError(
                "Please, specify the graph path (-g <path>) for a 'local' rpc instance"
            )
        # Only required when spawning a local rpc instance
        config["graph"]["grpc_server"]["path"] = graph_path
    elif cls == "remote":
        if "grpc_server" not in config_graph:
            raise ValueError(
                'Please, specify the "grpc_server" key configuration in the "graph" section'
            )
    else:
        raise ValueError(
            f'Value for "cls" must be "local"/"local_rust" or "remote", not {cls!r}'
        )

    app = make_app(config=config)

    aiohttp.web.run_app(app, host=host, port=port)


@graph_cli_group.command(name="download")
@click.option(
    "--s3-url",
    default=None,
    help="S3 directory containing the graph to download. "
    "Defaults to '{s3_prefix}/{name}/compressed/'",
)
@click.option(
    "--s3-prefix",
    default="s3://softwareheritage/graph/",
    help="Base directory of Software Heritage's graphs on S3",
)
@click.option(
    "--name",
    default=None,
    help="Name of the dataset to download. This is an ISO8601 date, optionally with a "
    "suffix. See https://docs.softwareheritage.org/devel/swh-export/graph/dataset.html",
)
@click.option(
    "--parallelism",
    "-j",
    default=5,
    help="Number of threads used to download/decompress files.",
)
@click.argument(
    "target-dir",
    type=click.Path(
        file_okay=False,
        writable=True,
        path_type=Path,
    ),
)
@click.pass_context
def download(
    ctx,
    s3_url: Optional[str],
    s3_prefix: str,
    name: Optional[str],
    parallelism: int,
    target_dir: Path,
):
    """Downloads a compressed SWH graph to the given target directory"""
    from swh.graph.download import GraphDownloader

    if s3_url and name:
        raise click.ClickException("--s3-url and --name are mutually exclusive")
    elif not s3_url and not name:
        raise click.ClickException("Either --s3-url or --name must be provided")
    elif not s3_url:
        s3_url = f"{s3_prefix.rstrip('/')}/{name}/compressed/"

    target_dir.mkdir(parents=True, exist_ok=True)

    GraphDownloader(
        local_graph_path=target_dir,
        s3_graph_path=s3_url,
        parallelism=parallelism,
    ).download_graph(
        progress_percent_cb=lambda _: None,
        progress_status_cb=lambda _: None,
    )


@graph_cli_group.command(name="list-datasets")
@click.option(
    "--s3-bucket",
    default="softwareheritage",
    help="S3 bucket name containing Software Heritage graph datasets. "
    "Defaults to 'sotwareheritage'",
)
@click.pass_context
def list_datasets(
    ctx,
    s3_bucket: Optional[str],
):
    """List graph datasets available for download.

    Print the names of the Software Heritage graph datasets that can be
    downloaded with the following command:

    $ swh graph download --name <dataset_name> <target_directory>

    The list may contain datasets that are not suitable for production, or not yet
    fully available. See
    https://docs.softwareheritage.org/devel/swh-export/graph/dataset.html
    for the official list of datasets, along with release notes.
    """
    import boto3
    from botocore import UNSIGNED
    from botocore.client import Config

    click.echo(
        "The following list is automatically generated from Software Heritage's "
        "S3 bucket.",
        err=True,
    )
    click.echo(
        "It may contain datasets that are not suitable for production, or not yet "
        "fully available.",
        err=True,
    )
    click.echo(
        "See https://docs.softwareheritage.org/devel/swh-export/graph/dataset.html "
        "for the official list of datasets, along with release notes.",
        err=True,
    )

    s3_client = boto3.client("s3", config=Config(signature_version=UNSIGNED))
    paginator = s3_client.get_paginator("list_objects_v2")

    for dataset_prefix in [
        prefix["Prefix"]
        for page in paginator.paginate(Bucket=s3_bucket, Prefix="graph/", Delimiter="/")
        for prefix in page["CommonPrefixes"]
    ]:
        if "Contents" in s3_client.list_objects_v2(
            Bucket=s3_bucket,
            Prefix=dataset_prefix + "compressed/",
            Delimiter="/",
        ):
            click.echo(dataset_prefix.split("/")[1])


@graph_cli_group.command(name="reindex")
@click.option(
    "--force",
    is_flag=True,
    help="Regenerate files even if they already exist. Implies --ef",
)
@click.option(
    "--ef", is_flag=True, help="Regenerate .ef files even if they already exist"
)
@click.argument(
    "graph",
    type=click.Path(
        writable=True,
    ),
)
@click.pass_context
def reindex(ctx, force: bool, ef: bool, graph: str):
    """Reindex a SWH GRAPH to the latest graph format.

    GRAPH should be composed of the graph folder followed by the graph prefix
    (by default "graph") eg. "graph_folder/graph".
    """
    import os.path

    from swh.graph.shell import Rust

    ef = ef or force
    conf = ctx.obj["config"]
    if "profile" not in conf:
        conf["profile"] = "release"

    if (
        ef
        or not os.path.exists(f"{graph}.ef")
        or not os.path.exists(f"{graph}-transposed.ef")
    ):
        logger.info("Recreating Elias-Fano indexes on adjacency lists")
        Rust("swh-graph-index", "ef", f"{graph}", conf=conf).run()
        Rust("swh-graph-index", "ef", f"{graph}-transposed", conf=conf).run()

    if (
        ef
        or not os.path.exists(f"{graph}-labelled.ef")
        or not os.path.exists(f"{graph}-labelled-transposed.ef")
    ):
        with open(f"{graph}.nodes.count.txt", "rt") as f:
            node_count = f.read().strip()

        # ditto
        logger.info("Recreating Elias-Fano indexes on arc labels")
        Rust(
            "swh-graph-index",
            "labels-ef",
            f"{graph}-labelled",
            node_count,
            conf=conf,
        ).run()
        Rust(
            "swh-graph-index",
            "labels-ef",
            f"{graph}-transposed-labelled",
            node_count,
            conf=conf,
        ).run()

    node2type_fname = f"{graph}.node2type.bin"
    if force or not os.path.exists(node2type_fname):
        logger.info("Creating node2type.bin")
        if os.path.exists(node2type_fname):
            os.unlink(node2type_fname)
        Rust("swh-graph-node2type", graph, conf=conf).run()


@graph_cli_group.command(name="grpc-serve")
@click.option(
    "--port",
    "-p",
    default=None,
    type=click.INT,
    metavar="PORT",
    show_default=True,
    help=(
        "port to bind the server on (note: host is not configurable "
        "for now and will be 0.0.0.0). Defaults to 50091"
    ),
)
@click.option("--graph", "-g", metavar="GRAPH", help="compressed graph basename")
@click.pass_context
def grpc_serve(ctx, port, graph):
    """Run the compressed graph gRPC service.

    This command uses execve to execute the Rust GRPC service.

    The documentation of the gRPC API is available on
    https://docs.softwareheritage.org/devel/swh-graph/grpc-api.html
    """
    import os

    from swh.graph.grpc_server import build_rust_grpc_server_cmdline

    config = ctx.obj["config"]

    profile = ctx.obj["config"].get("profile", "release")
    config["graph"]["profile"] = profile

    if graph is not None:
        config["graph"]["path"] = graph
    else:
        if "path" not in config["graph"]:
            raise ValueError("No graph base name provided")

    if port is None:
        port = config["graph"].get("port", 50091)
    config["graph"]["port"] = port

    logger.debug("Building gPRC server command line")
    cmd, port = build_rust_grpc_server_cmdline(**config["graph"])

    rust_bin = cmd[0]

    # XXX: shlex.join() is in 3.8
    # logger.info("Starting gRPC server: %s", shlex.join(cmd))
    logger.info("Starting gRPC server: %s", " ".join(shlex.quote(x) for x in cmd))
    os.execvp(rust_bin, cmd)


@graph_cli_group.command()
@click.option(
    "--input-dataset",
    "-i",
    type=PathlibPath(),
    help="graph dataset directory, in ORC format",
)
@click.option(
    "--sensitive-input-dataset",
    type=PathlibPath(),
    help="graph sensitive dataset directory, in ORC format",
)
@click.option(
    "--output-directory",
    "-o",
    type=PathlibPath(),
    help="directory where to store compressed graph",
)
@click.option(
    "--sensitive-output-directory",
    type=PathlibPath(),
    help="directory where to store sensitive compressed graph data",
)
@click.option(
    "--graph-name",
    "-g",
    default="graph",
    metavar="NAME",
    help="name of the output graph (default: 'graph')",
)
@click.option(
    "--steps",
    "-s",
    metavar="STEPS",
    type=StepOption(),
    help="run only these compression steps (default: all steps)",
)
@click.option("--check-flavor", type=str, help="Check flavor")
@click.pass_context
def compress(
    ctx,
    input_dataset,
    output_directory,
    sensitive_input_dataset,
    sensitive_output_directory,
    graph_name,
    steps,
    check_flavor,
):
    """Compress a graph using WebGraph

    Input: a directory containing a graph dataset in ORC format

    Output: a directory containing a WebGraph compressed graph

    Compression steps are: (1) extract_nodes, (2) mph, (3) bv, (4) bfs, (5)
    permute_bfs, (6) transpose_bfs, (7) simplify, (8) llp, (9) permute_llp,
    (10) obl, (11) compose_orders, (12) stats, (13) transpose, (14)
    transpose_obl, (15) maps, (16) extract_persons, (17) mph_persons, (18)
    node_properties, (19) mph_labels, (20) fcl_labels, (21) edge_labels, (22)
    edge_labels_obl, (23) edge_labels_transpose_obl, (24) clean_tmp.
    Compression steps can be selected by name or number using --steps,
    separating them with commas; step ranges (e.g., 3-9, 6-, etc.) are also
    supported.

    """
    from swh.graph import webgraph

    try:
        conf = ctx.obj["config"]["graph"]["compress"]
    except KeyError:
        conf = {}  # use defaults

    try:
        conf["profile"] = ctx.obj["config"]["profile"]
    except KeyError:
        conf["profile"] = "release"  # use release builds by default

    if check_flavor is None:
        # TODO: see is this can be None
        check_flavor = conf.get("check_flavor", "full")
    conf["check_flavor"] = check_flavor

    try:
        webgraph.compress(
            graph_name,
            input_dataset,
            output_directory,
            sensitive_input_dataset,
            sensitive_output_directory,
            check_flavor,
            steps,
            conf,
        )
    except webgraph.CompressionSubprocessError as e:
        try:
            if e.log_path.is_file():
                with e.log_path.open("rb") as f:
                    if e.log_path.stat().st_size > 1000:
                        f.seek(-1000, 2)  # read only the last 1kB
                        f.readline()  # skip first line, might be partial
                        sys.stderr.write("[...]\n")
                    sys.stderr.write("\n")
                    sys.stderr.flush()
                    sys.stderr.buffer.write(f.read())
                    sys.stderr.flush()
        except Exception:
            raise
            pass
        raise click.ClickException(e.message)


def get_all_subclasses(cls):
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses


@graph_cli_group.command()
@click.option(
    "--base-directory",
    required=True,
    type=PathlibPath(),
    help="""The base directory where all datasets and compressed graphs are.
    Its subdirectories should be named after a date (and optional flavor).
    For example: ``/poolswh/softwareheritage/``.""",
)
@click.option(
    "--athena-prefix",
    required=False,
    type=str,
    help="""A prefix for the Athena Database that will be created and/or used.
    For example: ``swh``.""",
)
@click.option(
    "--s3-prefix",
    required=False,
    type=str,
    help="""The base S3 "directory" where all datasets and compressed graphs are.
    Its subdirectories should be named after a date (and optional flavor).
    For example: ``s3://softwareheritage/graph/``.""",
)
@click.option(
    "--max-ram",
    help="""Maximum RAM that some scripts will try not to exceed""",
)
@click.option(
    "--batch-size",
    type=int,
    help="""Default value for compression tasks handling objects in batch""",
)
@click.option(
    "--grpc-api", help="""Default value for the <hostname>:<port> of the gRPC server"""
)
@click.option(
    "--s3-athena-output-location",
    required=False,
    type=str,
    help="""The base S3 "directory" where all datasets and compressed graphs are.
    Its subdirectories should be named after a date (and optional flavor).
    For example: ``s3://softwareheritage/graph/``.""",
)
@click.option(
    "--graph-base-directory",
    required=False,
    type=PathlibPath(),
    help="""Overrides the path of the graph to use. Defaults to the value of
    ``{base_directory}/{dataset_name}/{compressed}/``.
    For example: ``/dev/shm/swh-graph/default/``.""",
)
@click.option(
    "--export-base-directory",
    required=False,
    type=PathlibPath(),
    help="""Overrides the path of the export to use. Defaults to the value of
    ``--base-directory``.""",
)
@click.option(
    "--dataset-name",
    required=True,
    type=str,
    help="""Should be a date and optionally a flavor, which will be used
    as directory name. For example: ``2022-04-25`` or ``2022-11-12_staging``.""",
)
@click.option(
    "--parent-dataset-name",
    required=False,
    type=str,
    help="""When generating a subdataset (eg. ``2024-08-23-python3k``), this is the name
    of a full export (eg. ``2024-08-23``) the subdataset should be built from.""",
)
@click.option(
    "--export-name",
    required=False,
    type=str,
    help="""Should be a date and optionally a flavor, which will be used
    as directory name for the export (not the compressed graph).
    For example: ``2022-04-25`` or ``2022-11-12_staging``.
    Defaults to the value of --dataset-name""",
)
@click.option(
    "--parent-export-name",
    required=False,
    type=str,
    help="""When generating a subdataset (eg. ``2024-08-23-python3k``), this is the name
    of a full export (eg. ``2024-08-23``) the subdataset should be built from.
    Defaults to the value of --parent-dataset-name""",
)
@click.option(
    "--luigi-config",
    type=PathlibPath(),
    help="""Extra options to add to ``luigi.cfg``, following the same format.
    This overrides any option that would be other set automatically.""",
)
@click.option(
    "--retry-luigi-delay",
    type=int,
    default=10,
    help="""Time to wait before re-running Luigi, if some tasks are pending
    but stuck.""",
)
@click.argument("luigi_param", nargs=-1)
@click.pass_context
def luigi(
    ctx,
    base_directory: Path,
    graph_base_directory: Optional[Path],
    export_base_directory: Optional[Path],
    s3_prefix: Optional[str],
    athena_prefix: Optional[str],
    max_ram: Optional[str],
    batch_size: Optional[int],
    grpc_api: Optional[str],
    s3_athena_output_location: Optional[str],
    dataset_name: str,
    parent_dataset_name: Optional[str],
    parent_export_name: Optional[str],
    export_name: Optional[str],
    retry_luigi_delay: int,
    luigi_config: Optional[Path],
    luigi_param: List[str],
):
    r"""
    Internal command of swh-graph. Use 'swh export luigi' instead.

    Calls Luigi with the given task and params, and automatically
    configures paths based on --base-directory and --dataset-name.

    The list of Luigi params should be prefixed with ``--`` so they are not interpreted
    by the ``swh`` CLI. For example::

        swh datasets luigi \
                --base-directory ~/tmp/ \
                --dataset-name 2022-12-05_test \
                -- \
                RunAll \
                --local-scheduler

    to pass ``RunAll --local-scheduler`` as Luigi params

    Or, to compute a derived dataset::

        swh graph luigi \
                --graph-base-directory /dev/shm/swh-graph/default/ \
                --base-directory /poolswh/softwareheritage/vlorentz/ \
                --athena-prefix swh \
                --dataset-name 2022-04-25 \
                --s3-athena-output-location s3://some-bucket/tmp/athena \
                -- \
                --log-level INFO \
                FindEarliestRevisions \
                --scheduler-url http://localhost:50092/ \
                --blob-filter citation
    """
    import configparser
    import os
    import secrets
    import socket
    import subprocess
    import tempfile
    import time

    import luigi
    import psutil

    from swh.core.config import merge_configs

    # Popular the list of subclasses of luigi.Task
    import swh.export.luigi  # noqa
    from swh.graph.config import check_config, check_config_compress
    import swh.graph.luigi  # noqa

    config = configparser.ConfigParser()

    # By default, Luigi always returns code 0.
    # See https://luigi.readthedocs.io/en/stable/configuration.html#retcode
    config["retcode"] = {
        "already_running": "128",
        "missing_data": "129",
        "not_run": "130",
        "task_failed": "131",
        "scheduling_error": "132",
        "unhandled_exception": "133",
    }

    if max_ram:
        if max_ram.endswith("G"):
            max_ram_mb = int(max_ram[:-1]) * 1024
        elif max_ram.endswith("M"):
            max_ram_mb = int(max_ram[:-1])
        else:
            raise click.ClickException(
                "--max-ram must be an integer followed by M or G"
            )
    else:
        max_ram_mb = psutil.virtual_memory().total // (1024 * 1024)

    # Only used by the local scheduler; otherwise it needs to be configured
    # in luigid.cfg
    hostname = socket.getfqdn()
    config["resources"] = {
        f"{hostname}_ram_mb": str(max_ram_mb),
    }

    base_rust_executable_dir = ctx.obj["config"].get("base_rust_executable_dir")
    swh_config = check_config(ctx.obj["config"], base_rust_executable_dir)
    if "compress" in swh_config.get("graph", {}):
        swh_config = check_config_compress(
            swh_config,
            graph_name=swh_config["graph"].get("path"),
            in_dir=swh_config["graph"]["compress"].get("in_dir"),
            out_dir=swh_config["graph"]["compress"].get("out_dir"),
            sensitive_in_dir=swh_config["graph"]["compress"].get("sensitive_in_dir"),
            sensitive_out_dir=swh_config["graph"]["compress"].get("sensitive_out_dir"),
            check_flavor=swh_config["graph"]["compress"].get("check_flavor"),
        )

    export_name = export_name or dataset_name
    parent_export_name = parent_export_name or parent_dataset_name

    export_path = (export_base_directory or base_directory) / export_name
    dataset_path = base_directory / dataset_name

    default_values = dict(
        local_export_path=export_path,
        local_graph_path=dataset_path / "compressed",
        derived_datasets_path=dataset_path,
        topological_order_dir=dataset_path / "topology/",
        origin_urls_path=dataset_path / "origin_urls.csv.zst",
        export_id=f"{export_name}-{secrets.token_hex(10)}",
        export_name=export_name,
        dataset_name=dataset_name,
    )

    default_values = merge_configs(default_values, swh_config)

    if graph_base_directory:
        default_values["local_graph_path"] = graph_base_directory

    if s3_prefix:
        default_values["s3_export_path"] = f"{s3_prefix.rstrip('/')}/{export_name}"
        default_values["s3_graph_path"] = (
            f"{s3_prefix.rstrip('/')}/{dataset_name}/compressed"
        )
        if parent_dataset_name:
            default_values["s3_parent_export_path"] = (
                f"{s3_prefix.rstrip('/')}/{parent_export_name}"
            )
            default_values["s3_parent_graph_path"] = (
                f"{s3_prefix.rstrip('/')}/{parent_dataset_name}/compressed"
            )

    if s3_athena_output_location:
        default_values["s3_athena_output_location"] = s3_athena_output_location

    if max_ram:
        default_values["max_ram_mb"] = max_ram_mb

    if batch_size:
        default_values["batch_size"] = batch_size

    if grpc_api:
        default_values["grpc_api"] = grpc_api

    if athena_prefix:
        default_values["athena_db_name"] = (
            f"{athena_prefix}_{dataset_name.replace('-', '')}"
        )
        if parent_dataset_name:
            default_values["athena_parent_db_name"] = (
                f"{athena_prefix}_{parent_dataset_name.replace('-', '')}"
            )

    for task_cls in get_all_subclasses(luigi.Task):
        task_name = task_cls.__name__
        # If the task has an argument with one of the known name, add the default value
        # to its config.
        task_config = {
            arg_name: str(arg_value)
            for arg_name, arg_value in default_values.items()
            if hasattr(task_cls, arg_name)
        }
        if task_config:
            config[task_name] = task_config

    # If any config is provided, add it.
    # This may override default arguments configured above.
    if luigi_config is not None:
        config.read(luigi_config)

    with tempfile.NamedTemporaryFile(mode="w+t", prefix="luigi_", suffix=".cfg") as fd:
        config.write(fd)
        fd.flush()

        while True:
            start_time = time.time()
            proc = subprocess.run(
                [
                    "luigi",
                    "--module",
                    "swh.export.luigi",
                    "--module",
                    "swh.graph.luigi",
                    *luigi_param,
                ],
                env={
                    "LUIGI_CONFIG_PATH": fd.name,
                    **os.environ,
                },
            )

            # If all tasks are stuck, run another loop to wait for them to be unstuck.
            # The normal way to do this would be to set config["worker"]["keep_alive"]
            # to True, but we do it this way to force Luigi to recompute resources
            # because most tasks in compressed_graph.py can only know their resource
            # usage after ExtractNodes and ExtractPersons ran.
            if proc.returncode != int(config["retcode"]["not_run"]):
                break

            # wait a few seconds between loops to avoid wasting CPU
            time.sleep(max(0, retry_luigi_delay - (time.time() - start_time)))
        exit(proc.returncode)


@graph_cli_group.command(name="find-context")
@click.option(
    "-g",
    "--graph-grpc-server",
    default="localhost:50091",
    metavar="GRAPH_GRPC_SERVER",
    show_default=True,
    help="Graph RPC server address: as host:port",
)
@click.option(
    "-c",
    "--content-swhid",
    default="swh:1:cnt:3b997e8ef2e38d5b31fb353214a54686e72f0870",
    metavar="CNTSWHID",
    show_default=True,
    help="SWHID of the content",
)
@click.option(
    "-f",
    "--filename",
    default="",
    metavar="FILENAME",
    show_default=True,
    help="Name of file to search for",
)
@click.option(
    "-o",
    "--origin-url",
    default="",
    metavar="ORIGINURL",
    show_default=True,
    help="URL of the origin where we look for a content",
)
@click.option(
    "--all-origins/--no-all-origins",
    default=False,
    help="Compute fqswhid for all origins",
)
@click.option(
    "--fqswhid/--no-fqswhid",
    default=True,
    help="Compute fqswhid. If disabled, print only the origins.",
)
@click.option(
    "--trace/--no-trace",
    default=False,
    help="Print nodes examined while building fully qualified SWHID.",
)
@click.option(
    "--random-origin/--no-random-origin",
    default=True,
    help="Compute fqswhid for a random origin",
)
@click.pass_context
def find_context(ctx, **kwargs):
    """Utility to get the fully qualified SWHID for a given core SWHID.
    Uses the graph traversal to find the shortest path to an origin, and
    retains the first seen revision or release as anchor for cnt and dir types."""
    from swh.graph.find_context import main as lookup

    return lookup(**kwargs)


@graph_cli_group.command(name="link")
@click.argument(
    "source-path", type=click.Path(dir_okay=True, exists=True, path_type=Path)
)
@click.argument(
    "destination-path", type=click.Path(dir_okay=True, writable=True, path_type=Path)
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Explain what is being done"
)
@click.argument("force-copy", type=click.Path(path_type=Path), nargs=-1)
@click.pass_context
def link(
    ctx,
    source_path: Path,
    destination_path: Path,
    force_copy: Tuple[Path],
    verbose: bool,
):
    """
    Symlink (or copy) an existing graph to the desired location.

    By default, all files are symlinked, but files and directories can be
    specified to be copied instead.

    This functionality is intended for internal use, and is there to ease the
    process of sharing an existing graph between multiple users on the same
    machine.
    """
    import shutil

    from tqdm.contrib.concurrent import thread_map

    destination_path.mkdir(parents=True, exist_ok=True)

    for file_or_dir in source_path.rglob("*"):
        if file_or_dir in force_copy:
            continue
        link_target = destination_path / file_or_dir.relative_to(source_path)
        if file_or_dir.is_dir():
            link_target.mkdir(parents=True, exist_ok=True)
            continue
        link_target.symlink_to(file_or_dir)
        if verbose:
            logger.info(f"Creating symlink from {file_or_dir} to {link_target}")

    def _copy(source_item: Path):
        if verbose:
            logger.info(f"Copying {source_item} to {destination_path}")
        if source_item.is_file():
            shutil.copy(source_item, destination_path)
        else:
            shutil.copytree(source_item, destination_path)

    if len(force_copy) > 0:
        thread_map(_copy, force_copy, desc="Copying files", max_workers=len(force_copy))


def main():
    return graph_cli_group(auto_envvar_prefix="SWH_GRAPH")


if __name__ == "__main__":
    main()



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

r"""This module implements a shell-like command pipeline system in
pure-Python.

Pipelines are built like this:

>>> from swh.graph.shell import Command, Sink
>>> (
...     Command.echo("foo")
...     | Command.zstdmt()
...     | Command.cat("-", Command.echo("bar") | Command.zstdmt())
...     | Command.zstdcat()
...     > Sink()
... ).run().stdout
b'foo\nbar\n'

which is the equivalent of this bash command:

.. code-block:: bash

    echo foo \
    | zstdmt \
    | cat - <(echo bar | zstdmt) \
    | zstdcat

:class:`Sink` is mainly meant for tests; it causes ``.run().stdout`` to
return the stdout of the last process.

Actual pipelines will usually write to a file instead, using
:class:`AtomicFileSink`. This calls is similar to ``>`` in bash,
with a twist: it is only written after all other commands in the pipeline
succeeded (but unlike ``sponge`` from moreutils, it buffers to disk and
rename the file at the end).
"""

from __future__ import annotations

import dataclasses
import functools
import logging
import os
from pathlib import Path
import shlex
import signal
import subprocess
from typing import Any, BinaryIO, Dict, List, Optional, Tuple, TypeVar, Union

try:
    import luigi
    from luigi import LocalTarget
except ImportError:

    class LocalTarget:  # type: ignore
        """Placeholder for ``luigi.LocalTarget`` if it could not be imported"""

        pass


logger = logging.getLogger(__name__)


LOGBACK_CONF = b"""\
<configuration>
  <appender name="STDERR" class="ch.qos.logback.core.ConsoleAppender">
    <target>System.err</target>
    <encoder>
      <pattern>%d{HH:mm:ss.SSS} [%thread] %-5level %logger{36} %msg%n</pattern>
    </encoder>
  </appender>

  <root level="debug">
    <appender-ref ref="STDERR" />
  </root>
</configuration>
"""
"""Overrides the default config, to log to stderr instead of stdout"""


class CommandException(Exception):
    def __init__(self, command, returncode):
        super().__init__(f"{command[0]} returned: {returncode}")
        self.command = command
        self.returncode = returncode


_PROC = Path("/proc/")
""":file:`/proc/`"""
_CGROUP_ROOT = Path("/sys/fs/cgroup/")
"""Base path of the cgroup filesystem"""


@functools.lru_cache(1)
def base_cgroup() -> Optional[Path]:
    """Returns the cgroup that should be used as parent for child processes.

    As `cgroups with children should not contain processes themselves
    <https://systemd.io/CGROUP_DELEGATION/#two-key-design-rules>`_, this is the parent
    of the cgroup this process was started in.
    """
    import atexit

    if not _CGROUP_ROOT.is_dir():
        logger.info("%s is not mounted", _CGROUP_ROOT)
        return None

    proc_cgroup_path = _PROC / str(os.getpid()) / "cgroup"
    if not proc_cgroup_path.is_file():
        logger.info("%s does not exist", proc_cgroup_path)
        return None

    my_cgroup = proc_cgroup_path.read_text().strip()
    if not my_cgroup.startswith("0::/"):
        # https://www.kernel.org/doc/html/latest/admin-guide/cgroup-v2.html#processes
        logger.warning("Process was started in %s which is not a cgroupv2", my_cgroup)
        return None

    # this is the cgroup that contains the current process, plus whatever process
    # spawned it (eg. pytest or bash); and neither cgroupv2 nor systemd allows a cgroup
    # to both contain processes itself and have child cgroups; so we have to use the
    # parent cgroup as root for the cgroups we are going to create.
    original_cgroup_path = _CGROUP_ROOT / my_cgroup[4:]
    if original_cgroup_path == _CGROUP_ROOT:
        # Running directly in the root cgroup, so there is no parent.
        # TODO: this means we are running in a container, so this is probably the only
        # process in the cgroup, we could try moving it to a child cgroup.
        return None
    assert (original_cgroup_path.parent / "cgroup.procs").read_text().strip() == ""

    # create a cgroup that will encapsulate both the "swh.graph.shell" cgroup and
    # all the children
    base_cgroup_path = create_cgroup(
        f"swh.graph@{os.getpid()}", original_cgroup_path.parent, add_suffix=False
    )
    if base_cgroup_path is None:
        return None

    assert (base_cgroup_path / "cgroup.procs").read_text().strip() == ""
    for controller in ("cpu", "memory"):
        try:
            with (base_cgroup_path / "cgroup.subtree_control").open("wt") as f:
                f.write(f"+{controller}\n")
        except OSError as e:
            logger.warning(
                "Failed to enable %r controller for %s: %s",
                controller,
                base_cgroup_path,
                e,
            )

    def cleanup():
        # Clean up the base cgroup we created
        base_cgroup_path.rmdir()

    atexit.register(cleanup)

    return base_cgroup_path


_num_child_cgroups = 0


def create_cgroup(
    base_name: str, parent: Optional[Path] = None, add_suffix: bool = True
) -> Optional[Path]:
    global _num_child_cgroups

    parent = parent or base_cgroup()
    if parent is None:
        return None

    if add_suffix:
        name = f"{base_name}@{_num_child_cgroups}"
        _num_child_cgroups += 1
    else:
        name = base_name

    new_cgroup_path = parent / name
    try:
        new_cgroup_path.mkdir()
    except OSError as e:
        logger.warning("Failed to create %s: %s", new_cgroup_path, e)
        return None

    return new_cgroup_path


def move_to_cgroup(cgroup: Path, pid: Optional[int] = None) -> bool:
    """Returns whether the process was successfully moved."""
    if pid is None:
        pid = os.getpid()
    try:
        with (cgroup / "cgroup.procs").open("at") as f:
            f.write(f"{pid}\n")
    except OSError as e:
        logger.warning("Failed to move process to %s: %s", cgroup, e)
        cgroup.rmdir()
        return False
    else:
        return True


class _MetaCommand(type):
    def __getattr__(self, name):
        return functools.partial(Command, name)


class Command(metaclass=_MetaCommand):
    """Runs a command with the given name and arguments. ``**kwargs`` is passed to
    :class:`subprocess.Popen`.

    If ``check`` is :const:`True` (the default), raises an exception if the command
    returns a non-zero exit code."""

    def __init__(
        self, *args: Union[str, Path, LocalTarget], check: bool = True, **kwargs
    ):
        self.args = args
        self.kwargs = dict(kwargs)
        self.preexec_fn = self.kwargs.pop("preexec_fn", lambda: None)
        self.cgroup = None
        self.check = check

    def _preexec_fn(self):
        if self.cgroup is not None:
            move_to_cgroup(self.cgroup)
        self.preexec_fn()

    def _run(self, stdin, stdout, stderr) -> _RunningCommand:
        cgroup = create_cgroup(str(self.args[0]).split("/")[-1])
        pass_fds = []
        children = []
        final_args = []
        pipes_to_close: List[int] = []
        stdout_read = None
        stderr_read = None

        if stderr is subprocess.STDOUT:
            if stdout is subprocess.PIPE:
                (r, w) = os.pipe()
                stdout = w
                stderr = w
                stdout_read = r
                pipes_to_close.append(w)
            else:
                stderr = stdout
        elif stderr is subprocess.PIPE:
            raise NotImplementedError(
                "swh.graph.shell.Command._run(..., stderr=subprocess.PIPE)"
            )

        for arg in self.args:
            if isinstance(arg, (Command, Pipe)):
                # command stdout piped to a non-stdin FD
                (r, w) = os.pipe()
                pass_fds.append(r)
                final_args.append(f"/dev/fd/{r}")
                children.append(arg._run(None, w, stderr))
                os.close(w)
            elif isinstance(arg, LocalTarget):
                final_args.append(arg.path)
            elif isinstance(arg, Path):
                final_args.append(str(arg))
            else:
                final_args.append(arg)

        proc = subprocess.Popen(
            final_args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            pass_fds=pass_fds,
            preexec_fn=self._preexec_fn,
            **self.kwargs,
        )

        for pipe in pipes_to_close:
            os.close(pipe)

        return _RunningCommand(
            self,
            proc,
            children,
            cgroup,
            stdout=(
                os.fdopen(stdout_read, "rb")
                if isinstance(stdout_read, int)
                else stdout_read
            ),
            stderr=(
                os.fdopen(stderr_read, "rb")
                if isinstance(stderr_read, int)
                else stderr_read
            ),
            check=self.check,
        )

    def run(self, stderr=None) -> List[RunResult]:
        return self._run(None, None, stderr).wait()

    def __or__(self, other: Union[Command, Pipe]) -> Pipe:
        """``self | other``: pipe self's stdout to other's stdin"""
        if isinstance(other, Command):
            return Pipe([self, other])
        elif isinstance(other, Pipe):
            return Pipe([self, *other.children])
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__} | {other.__class__.__name__}"
            )

    def __str__(self) -> str:
        return f"{' '.join(shlex.quote(str(arg)) for arg in self.args)}"


class Rust(Command):
    def __init__(
        self,
        bin_name,
        *args: Union[str, Path, "luigi.LocalTarget"],
        base_rust_executable_dir: Optional[Path] = None,
        conf: Optional[Dict[str, Any]] = None,
        env: Optional[Dict[str, str]] = None,
    ):
        from .config import check_config

        conf = dict(conf or {})
        conf = check_config(conf, base_rust_executable_dir=base_rust_executable_dir)
        assert conf is not None  # for mypy

        env = env or dict(os.environ)
        path = env.get("PATH")
        if path:
            env["PATH"] = f"{conf['rust_executable_dir']}:{path}"
        else:
            env["PATH"] = conf["rust_executable_dir"]
        env["RUST_MIN_STACK"] = "8388608"  # 8MiB; avoids stack overflows in LLP

        super().__init__(bin_name, *args, env=env)


class _RunningCommand:
    def __init__(
        self,
        command: Command,
        proc: subprocess.Popen,
        running_children: List[Union[_RunningCommand, _RunningPipe]],
        cgroup: Optional[Path],
        stdout: Optional[BinaryIO] = None,
        stderr: Optional[BinaryIO] = None,
        check: bool = True,
    ):
        self.command = command
        self.proc = proc
        self.running_children = running_children
        self.cgroup = cgroup
        self.check = check
        self._stdout = stdout
        self._stderr = stderr

    def stdout(self):
        return self._stdout or self.proc.stdout

    def stderr(self):
        return self._stderr or self.proc.stderr

    def is_alive(self) -> bool:
        return self.proc.poll() is None

    def wait(self) -> List[RunResult]:
        results = []
        try:
            self.proc.wait()
            results.append(
                RunResult(
                    command=tuple(map(str, self.command.args)),
                    cgroup=self.cgroup,
                    cgroup_stats={
                        p.name: p.read_text().strip()
                        for p in (self.cgroup.iterdir() if self.cgroup else [])
                        if p.name.startswith(("cpu.", "memory.", "io.", "pids."))
                        # exclude writeable files (they are for control, not statistics)
                        and p.stat().st_mode & 0o600 == 0o400
                    },
                )
            )
            self._cleanup()
            if self.check and self.proc.returncode not in (0, -int(signal.SIGPIPE)):
                raise CommandException(self.command.args, self.proc.returncode)

            for child in self.running_children:
                results.extend(child.wait())
        except BaseException:
            self.kill()
            raise

        return results

    def kill(self) -> None:
        for child in self.running_children:
            child.kill()

        if self.proc.returncode is not None:
            self.proc.kill()

    def _cleanup(self) -> None:
        if self.cgroup is not None:
            self.cgroup.rmdir()


class Pipe:
    def __init__(self, children: List[Union[Command, Pipe]]):
        self.children = children

    def _run(self, stdin, stdout, stderr) -> _RunningPipe:
        read_pipes: List[Any] = [stdin]
        write_pipes: List[Any] = []
        pipes_to_close: List[int] = []
        for _ in range(len(self.children) - 1):
            (r, w) = os.pipe()
            read_pipes.append(r)
            write_pipes.append(w)
            pipes_to_close.append(r)
            pipes_to_close.append(w)

        if stdout is subprocess.PIPE:
            (r, w) = os.pipe()
            stdout = w
            stdout_read = os.fdopen(r, "rb")
            pipes_to_close.append(w)
        elif stdout is subprocess.DEVNULL:
            stdout_read = None
        else:
            stdout_read = os.fdopen(stdout, "rb") if isinstance(stdout, int) else stdout

        write_pipes.append(stdout)

        if stderr is subprocess.STDOUT:
            stderr = stdout
        elif stderr is subprocess.PIPE:
            raise NotImplementedError(
                "swh.graph.shell.Pipe._run(..., stderr=subprocess.PIPE)"
            )

        running_children = [
            child._run(r, w, stderr)
            for (r, w, child) in zip(read_pipes, write_pipes, self.children)
        ]

        # We need to close these file descriptors in the parent process, so that the
        # corresponding pipes have an end fully closed when the corresponding process
        # dies.
        # On CPython (and when not running in tools like pytest that keep references to
        # call frames), this would be done automatically when the function exits because
        # the refcount to all pipes becomes 0, but we cannot rely
        # on this behavior.
        # Without it, the other end of the process will hang when trying to read
        # or write to it.
        for pipe in pipes_to_close:
            os.close(pipe)

        return _RunningPipe(self, running_children, stdout_read)

    def run(self) -> List[RunResult]:
        return self._run(None, None, None).wait()

    def __or__(self, other) -> Pipe:
        if isinstance(other, Pipe):
            return Pipe([*self.children, *other.children])
        elif isinstance(other, Command):
            return Pipe([*self.children, other])
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__} | {other.__class__.__name__}"
            )

    def __str__(self) -> str:
        children = "\n| ".join(map(str, self.children))
        return f"( {children}\n)"


def wc(source: Union[Command, Pipe], *args: str) -> int:
    return int((source | Command.wc(*args) > Sink()).run().stdout.strip())


class _RunningPipe:
    def __init__(
        self,
        pipe: Pipe,
        children: List[Union[_RunningCommand, _RunningPipe]],
        stdout: Optional[BinaryIO],
    ):
        self.pipe = pipe
        self.children = children
        self._stdout = stdout

    def stdout(self):
        return self._stdout

    def is_alive(self) -> bool:
        return all(child.is_alive() for child in self.children)

    def wait(self) -> List[RunResult]:
        results = []
        try:
            for child in self.children:
                results.extend(child.wait())
        except BaseException:
            self.kill()
            raise

        return results

    def kill(self) -> None:
        for child in self.children:
            child.kill()


TSink = TypeVar("TSink", bound="_BaseSink")


class _BaseSink:
    def __init__(self) -> None:
        self.source_pipe: Union[None, Command, Pipe] = None

    def _run(self, stdin, stdout, stderr) -> Any:
        raise TypeError(f"{self.__class__.__name__} must be the end of a pipeline.")

    def __lt__(self: TSink, other: Union[Command, Pipe]) -> TSink:
        """``other > self``"""
        if isinstance(other, (Command, Pipe)):
            if self.source_pipe is not None:
                raise TypeError(f"{self!r} is already piped to {self.source_pipe!r}")
            self.source_pipe = other
            return self
        else:
            raise NotImplementedError(
                f"{other.__class__.__name__} > {self.__class__.__name__}"
            )


class _SinkResult:
    def __init__(self, result: List[RunResult], stdout: bytes) -> None:
        self.result = result
        self.stdout = stdout


class Sink(_BaseSink):
    """Captures the final output instead of sending it to the process' stdout"""

    def run(self, stderr=None) -> _SinkResult:
        if self.source_pipe is None:
            raise TypeError("AtomicFileSink has no stdin")
        if stderr is subprocess.STDOUT:
            raise NotImplementedError(
                "swh.graph.shell.Sink.run(stderr=subprocess.STDOUT)"
            )

        source = self.source_pipe._run(
            stdin=None, stdout=subprocess.PIPE, stderr=stderr
        )

        chunks = []
        while True:
            new_chunk = source.stdout().read(10240)
            if not new_chunk and not source.is_alive():
                break
            chunks.append(new_chunk)

        return _SinkResult(result=source.wait(), stdout=b"".join(chunks))


class _RunningAtomicFileSink:
    """Properly distributes ``stdout`` and ``stderr`` during ``AtomicFileSink``."""

    def __init__(
        self,
        source_pipe,
        stdout: Optional[BinaryIO],
        stderr: Optional[BinaryIO],
        path,
        tmp_path,
        tmp_fd,
    ):
        self.source_pipe = source_pipe
        self._stdout = stdout
        self._stderr = stderr
        self.path = path
        self.tmp_path = tmp_path
        self.tmp_fd = tmp_fd

    def stdout(self):
        return self._stdout

    def stderr(self):
        return self._stderr

    def wait(self):
        try:
            result = self.source_pipe.wait()
        except BaseException:
            self.tmp_fd.close()
            self.tmp_path.unlink()
            raise
        else:
            self.tmp_fd.close()
            self.tmp_path.replace(self.path)
            return result


class AtomicFileSink(_BaseSink):
    """Similar to ``> path`` at the end of a command, but writes only if the whole
    command succeeded."""

    def __init__(self, path: Union[Path, LocalTarget]):
        super().__init__()
        if isinstance(path, LocalTarget):
            path = Path(path.path)
        self.path = path

    def _run(self, stdin, stdout, stderr) -> _RunningAtomicFileSink:
        if self.source_pipe is None:
            raise TypeError("AtomicFileSink has no stdin")

        tmp_path = Path(f"{self.path}.tmp")
        if tmp_path.exists():
            tmp_path.unlink()
        tmp_fd = tmp_path.open("wb")

        pipes_to_close: List[int] = []

        if stderr is subprocess.STDOUT:
            (r, w) = os.pipe()
            stdout = os.fdopen(r, "rb")
            stderr_read = stdout
            stderr = w
            pipes_to_close.append(w)
        elif stderr is subprocess.PIPE:
            (r, w) = os.pipe()
            stdout = None
            stderr_read = os.fdopen(r, "rb")
            stderr = w
            pipes_to_close.append(w)
        else:
            stdout = None
            stderr_read = None

        running_sink = _RunningAtomicFileSink(
            source_pipe=self.source_pipe._run(stdin=None, stdout=tmp_fd, stderr=stderr),
            stdout=stdout,
            stderr=stderr_read,
            path=self.path,
            tmp_path=tmp_path,
            tmp_fd=tmp_fd,
        )

        for pipe in pipes_to_close:
            os.close(pipe)

        return running_sink

    def run(self, stderr=None) -> List[RunResult]:
        return self._run(None, None, stderr).wait()

    def __str__(self) -> str:
        return f"{self.source_pipe} > AtomicFileSink({self.path})"


@dataclasses.dataclass
class RunResult:
    cgroup: Optional[Path]
    command: Tuple[str, ...]
    cgroup_stats: Dict[str, str]



# Copyright (c) 2022-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib
import logging

import pytest
from pytest import raises

from swh.core.api import RemoteException
from swh.graph.http_client import GraphArgumentException

TEST_ORIGIN_ID = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph").hexdigest()
)


def test_remote_graph_client_with_bad_url(remote_graph_client_url):
    from swh.graph.http_client import RemoteGraphClient

    with pytest.raises(ValueError, match="URL is incorrect"):
        _ = RemoteGraphClient(f"{remote_graph_client_url}/GARBAGE/")


def test_remote_graph_client_with_service_down():
    from swh.graph.http_client import GraphAPIError, RemoteGraphClient

    with pytest.raises(GraphAPIError):
        _ = RemoteGraphClient("http://localhost:9/graph/")


def test_remote_graph_client_errors_on_unexpected_stats(
    mocker, remote_graph_client_url
):
    from swh.graph.http_client import RemoteGraphClient

    mocker.patch.object(RemoteGraphClient, "stats", return_value={})
    with pytest.raises(ValueError, match="unexpected results"):
        _ = RemoteGraphClient(remote_graph_client_url)


def test_remote_graph_client_log_export_started_at(
    mocker, caplog, remote_graph_client_url
):
    from swh.graph.http_client import RemoteGraphClient

    caplog.set_level(logging.DEBUG, logger="swh.graph.http_client")
    mocker.patch.object(
        RemoteGraphClient,
        "stats",
        return_value={"num_nodes": 42, "export_started_at": 1700000000},
    )
    _ = RemoteGraphClient(remote_graph_client_url)
    assert "Graph export started at 2023-11-14T22:13:20+00:00 (42 nodes)" in [
        rec.message for rec in caplog.records
    ]


def test_stats(graph_client):
    stats = graph_client.stats()
    assert stats["num_nodes"] == 24
    assert stats["num_edges"] == 28
    assert isinstance(stats["compression_ratio"], float)
    assert isinstance(stats["bits_per_node"], float)
    assert isinstance(stats["bits_per_edge"], float)
    assert stats["indegree_min"] == 0
    assert stats["indegree_max"] == 4
    assert isinstance(stats["indegree_avg"], float)
    assert stats["outdegree_min"] == 0
    assert stats["outdegree_max"] == 3
    assert isinstance(stats["outdegree_avg"], float)
    assert stats["export_started_at"] == 1669888200
    assert stats["export_ended_at"] == 1669899600


def test_leaves(graph_client, graph_grpc_backend_implementation):
    actual = list(graph_client.leaves(TEST_ORIGIN_ID))
    expected = [
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000007",
    ]
    assert set(actual) == set(expected)


@pytest.mark.parametrize("max_matching_nodes", [0, 1, 2, 3, 4, 5, 10, 1 << 31])
def test_leaves_with_limit(
    graph_client, max_matching_nodes, graph_grpc_backend_implementation
):
    actual = list(
        graph_client.leaves(TEST_ORIGIN_ID, max_matching_nodes=max_matching_nodes)
    )
    expected = [
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000007",
    ]

    if max_matching_nodes == 0:
        assert set(actual) == set(expected)
    else:
        assert set(actual) <= set(expected)
        assert len(actual) == min(4, max_matching_nodes)


def test_empty_neighbors(graph_client, graph_grpc_backend_implementation):
    # https://gitlab.softwareheritage.org/swh/devel/swh-graph/-/issues/4790
    actual = list(
        graph_client.neighbors(
            "swh:1:rev:0000000000000000000000000000000000000009",
            direction="forward",
            # We are guaranteed no results because no ori→snp edges exist from a rev
            edges="ori:snp",
        )
    )
    assert actual == []


def test_neighbors(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.neighbors(
            "swh:1:rev:0000000000000000000000000000000000000009", direction="backward"
        )
    )
    expected = [
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:snp:0000000000000000000000000000000000000020",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000013",
    ]
    assert set(actual) == set(expected)


def test_neighbors_max_edges2(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.neighbors(
            "swh:1:rev:0000000000000000000000000000000000000009",
            direction="backward",
            max_edges=2,
        )
    )
    expected = [
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:snp:0000000000000000000000000000000000000020",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000013",
    ]
    assert set(actual) <= set(expected)
    assert len(set(actual)) == 2


def test_neighbors_max_edges1(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.neighbors(
            "swh:1:rev:0000000000000000000000000000000000000009",
            direction="backward",
            max_edges=1,
        )
    )
    expected = [
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:snp:0000000000000000000000000000000000000020",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000013",
    ]
    assert set(actual) <= set(expected)
    assert len(set(actual)) == 1


def test_visit_nodes(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.visit_nodes(
            "swh:1:rel:0000000000000000000000000000000000000010",
            edges="rel:rev,rev:rev",
        )
    )
    expected = [
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
    ]
    assert set(actual) == set(expected)


@pytest.mark.parametrize("max_matching_nodes", [0, 1, 2, 3, 4, 5, 10, 1 << 31])
def test_visit_nodes_limit(
    graph_client, max_matching_nodes, graph_grpc_backend_implementation
):
    actual = list(
        graph_client.visit_nodes(
            "swh:1:rel:0000000000000000000000000000000000000010",
            edges="rel:rev,rev:rev",
            max_matching_nodes=max_matching_nodes,
        )
    )
    expected = [
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
    ]
    if max_matching_nodes == 0:
        assert set(actual) == set(expected)
    else:
        assert set(actual) <= set(expected)
        assert len(actual) == min(3, max_matching_nodes)


def test_visit_nodes_filtered(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.visit_nodes(
            "swh:1:rel:0000000000000000000000000000000000000010",
            return_types="dir",
        )
    )
    expected = [
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000006",
    ]
    assert set(actual) == set(expected)


@pytest.mark.parametrize("max_matching_nodes", [0, 1, 2, 3, 4, 5, 10, 1 << 31])
def test_visit_nodes_filtered_limit(
    graph_client, max_matching_nodes, graph_grpc_backend_implementation
):
    actual = list(
        graph_client.visit_nodes(
            "swh:1:rel:0000000000000000000000000000000000000010",
            return_types="dir",
            max_matching_nodes=max_matching_nodes,
        )
    )
    expected = [
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000006",
    ]
    if max_matching_nodes == 0:
        assert set(actual) == set(expected)
    else:
        assert set(actual) <= set(expected)
        assert len(actual) == min(3, max_matching_nodes)


def test_visit_nodes_filtered_star(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.visit_nodes(
            "swh:1:rel:0000000000000000000000000000000000000010",
            return_types="*",
        )
    )
    expected = [
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:cnt:0000000000000000000000000000000000000007",
        "swh:1:dir:0000000000000000000000000000000000000006",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
    ]
    assert set(actual) == set(expected)


def test_visit_edges(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.visit_edges(
            "swh:1:rel:0000000000000000000000000000000000000010",
            edges="rel:rev,rev:rev,rev:dir",
        )
    )
    expected = [
        (
            "swh:1:rel:0000000000000000000000000000000000000010",
            "swh:1:rev:0000000000000000000000000000000000000009",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:rev:0000000000000000000000000000000000000003",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000008",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000003",
            "swh:1:dir:0000000000000000000000000000000000000002",
        ),
    ]
    assert set(actual) == set(expected)


def test_visit_edges_limited(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.visit_edges(
            "swh:1:rel:0000000000000000000000000000000000000010",
            max_edges=4,
            edges="rel:rev,rev:rev,rev:dir",
        )
    )
    expected = [
        (
            "swh:1:rel:0000000000000000000000000000000000000010",
            "swh:1:rev:0000000000000000000000000000000000000009",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:rev:0000000000000000000000000000000000000003",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000008",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000003",
            "swh:1:dir:0000000000000000000000000000000000000002",
        ),
    ]

    # As there are four valid answers (up to reordering), we cannot check for
    # equality. Instead, we check the client returned either
    # * all edges but one, or
    # * all edges
    # and the right answer depends on which edges were traversed, which is
    # non-deterministic
    extras = set(actual) - set(expected)
    assert not extras, (
        "Returned unexpected arcs:\n    "
        + "\n    ".join(map(str, extras))
        + "\nExpected:\n    "
        + "\n    ".join(map(str, expected))
        + "\nGot:\n    "
        + "\n    ".join(map(str, actual))
    )
    assert 3 <= len(actual) <= 4


def test_visit_edges_diamond_pattern(graph_client, graph_grpc_backend_implementation):
    actual = list(
        graph_client.visit_edges(
            "swh:1:rev:0000000000000000000000000000000000000009",
            edges="*",
        )
    )
    expected = [
        (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:rev:0000000000000000000000000000000000000003",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000009",
            "swh:1:dir:0000000000000000000000000000000000000008",
        ),
        (
            "swh:1:rev:0000000000000000000000000000000000000003",
            "swh:1:dir:0000000000000000000000000000000000000002",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000002",
            "swh:1:cnt:0000000000000000000000000000000000000001",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000008",
            "swh:1:cnt:0000000000000000000000000000000000000001",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000008",
            "swh:1:cnt:0000000000000000000000000000000000000007",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000008",
            "swh:1:dir:0000000000000000000000000000000000000006",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000006",
            "swh:1:cnt:0000000000000000000000000000000000000004",
        ),
        (
            "swh:1:dir:0000000000000000000000000000000000000006",
            "swh:1:cnt:0000000000000000000000000000000000000005",
        ),
    ]
    assert set(actual) == set(expected)


@pytest.mark.skip(reason="currently disabled due to T1969")
def test_walk(graph_client, graph_grpc_backend_implementation):
    args = ("swh:1:dir:0000000000000000000000000000000000000016", "rel")
    kwargs = {
        "edges": "dir:dir,dir:rev,rev:*",
        "direction": "backward",
        "traversal": "bfs",
    }

    actual = list(graph_client.walk(*args, **kwargs))
    expected = [
        "swh:1:dir:0000000000000000000000000000000000000016",
        "swh:1:dir:0000000000000000000000000000000000000017",
        "swh:1:rev:0000000000000000000000000000000000000018",
        "swh:1:rel:0000000000000000000000000000000000000019",
    ]
    assert set(actual) == set(expected)

    kwargs2 = kwargs.copy()
    kwargs2["limit"] = -1
    actual = list(graph_client.walk(*args, **kwargs2))
    expected = ["swh:1:rel:0000000000000000000000000000000000000019"]
    assert set(actual) == set(expected)

    kwargs2 = kwargs.copy()
    kwargs2["limit"] = 2
    actual = list(graph_client.walk(*args, **kwargs2))
    expected = [
        "swh:1:dir:0000000000000000000000000000000000000016",
        "swh:1:dir:0000000000000000000000000000000000000017",
    ]
    assert set(actual) == set(expected)


@pytest.mark.skip(reason="Random walk is deprecated")
def test_random_walk_dst_is_type(graph_client, graph_grpc_backend_implementation):
    """as the walk is random, we test a visit from a cnt node to a release
    reachable from every single path in the backward graph, and only check the
    final node of the path (i.e., the release)
    """
    args = ("swh:1:cnt:0000000000000000000000000000000000000015", "rel")
    kwargs = {"direction": "backward"}
    expected_root = "swh:1:rel:0000000000000000000000000000000000000019"

    actual = list(graph_client.random_walk(*args, **kwargs))
    assert len(actual) > 1  # no release directly links to a content
    assert actual[0] == args[0]
    assert actual[-1] == expected_root

    kwargs2 = kwargs.copy()
    kwargs2["limit"] = -1
    actual = list(graph_client.random_walk(*args, **kwargs2))
    assert actual == [expected_root]

    kwargs2["limit"] = -2
    actual = list(graph_client.random_walk(*args, **kwargs2))
    assert len(actual) == 2
    assert actual[-1] == expected_root

    kwargs2["limit"] = 3
    actual = list(graph_client.random_walk(*args, **kwargs2))
    assert len(actual) == 3


@pytest.mark.skip(reason="Random walk is deprecated")
def test_random_walk_dst_is_node(graph_client, graph_grpc_backend_implementation):
    """Same as test_random_walk_dst_is_type, but we target the specific release
    node instead of a type
    """
    args = (
        "swh:1:cnt:0000000000000000000000000000000000000015",
        "swh:1:rel:0000000000000000000000000000000000000019",
    )
    kwargs = {"direction": "backward"}
    expected_root = "swh:1:rel:0000000000000000000000000000000000000019"

    actual = list(graph_client.random_walk(*args, **kwargs))
    assert len(actual) > 1  # no origin directly links to a content
    assert actual[0] == args[0]
    assert actual[-1] == expected_root

    kwargs2 = kwargs.copy()
    kwargs2["limit"] = -1
    actual = list(graph_client.random_walk(*args, **kwargs2))
    assert actual == [expected_root]

    kwargs2["limit"] = -2
    actual = list(graph_client.random_walk(*args, **kwargs2))
    assert len(actual) == 2
    assert actual[-1] == expected_root

    kwargs2["limit"] = 3
    actual = list(graph_client.random_walk(*args, **kwargs2))
    assert len(actual) == 3


def test_count(graph_client, graph_grpc_backend_implementation):
    actual = graph_client.count_leaves(TEST_ORIGIN_ID)
    assert actual == 4
    actual = graph_client.count_visit_nodes(
        "swh:1:rel:0000000000000000000000000000000000000010", edges="rel:rev,rev:rev"
    )
    assert actual == 3
    actual = graph_client.count_neighbors(
        "swh:1:rev:0000000000000000000000000000000000000009", direction="backward"
    )
    assert actual == 4


@pytest.mark.parametrize("max_matching_nodes", [0, 1, 2, 3, 4, 5, 10, 1 << 31])
def test_count_with_limit(
    graph_client, max_matching_nodes, graph_grpc_backend_implementation
):
    actual = graph_client.count_leaves(
        TEST_ORIGIN_ID, max_matching_nodes=max_matching_nodes
    )
    if max_matching_nodes == 0:
        assert actual == 4
    else:
        assert actual == min(4, max_matching_nodes)


def test_param_validation(graph_client, graph_grpc_backend_implementation):
    with raises(GraphArgumentException) as exc_info:  # SWHID not found
        list(graph_client.leaves("swh:1:rel:00ffffffff000000000000000000000000000010"))
    if exc_info.value.response:
        assert exc_info.value.response.status_code == 404

    with raises(GraphArgumentException) as exc_info:  # malformed SWHID
        list(
            graph_client.neighbors("swh:1:rel:00ffffffff00000000zzzzzzz000000000000010")
        )
    if exc_info.value.response:
        assert exc_info.value.response.status_code == 400

    with raises(GraphArgumentException) as exc_info:  # malformed edge specification
        list(
            graph_client.visit_nodes(
                "swh:1:dir:0000000000000000000000000000000000000016",
                edges="dir:notanodetype,dir:rev,rev:*",
                direction="backward",
            )
        )
    if exc_info.value.response:
        assert exc_info.value.response.status_code == 400

    with raises(GraphArgumentException) as exc_info:  # malformed direction
        list(
            graph_client.visit_nodes(
                "swh:1:dir:0000000000000000000000000000000000000016",
                edges="dir:dir,dir:rev,rev:*",
                direction="notadirection",
            )
        )
    if exc_info.value.response:
        assert exc_info.value.response.status_code == 400


@pytest.mark.skip(reason="currently disabled due to T1969")
def test_param_validation_walk(graph_client):
    """test validation of walk-specific parameters only"""
    with raises(RemoteException) as exc_info:  # malformed traversal order
        list(
            graph_client.walk(
                "swh:1:dir:0000000000000000000000000000000000000016",
                "rel",
                edges="dir:dir,dir:rev,rev:*",
                direction="backward",
                traversal="notatraversalorder",
            )
        )
    assert exc_info.value.response.status_code == 400



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from swh.graph.grpc import swhgraph_pb2 as swh_dot_graph_dot_grpc_dot_swhgraph__pb2

GRPC_GENERATED_VERSION = '1.69.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in swh/graph/grpc/swhgraph_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class TraversalServiceStub(object):
    """Graph traversal service 
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetNode = channel.unary_unary(
                '/swh.graph.TraversalService/GetNode',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.GetNodeRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Node.FromString,
                _registered_method=True)
        self.Traverse = channel.unary_stream(
                '/swh.graph.TraversalService/Traverse',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Node.FromString,
                _registered_method=True)
        self.FindPathTo = channel.unary_unary(
                '/swh.graph.TraversalService/FindPathTo',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.FindPathToRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Path.FromString,
                _registered_method=True)
        self.FindPathBetween = channel.unary_unary(
                '/swh.graph.TraversalService/FindPathBetween',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.FindPathBetweenRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Path.FromString,
                _registered_method=True)
        self.CountNodes = channel.unary_unary(
                '/swh.graph.TraversalService/CountNodes',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.CountResponse.FromString,
                _registered_method=True)
        self.CountEdges = channel.unary_unary(
                '/swh.graph.TraversalService/CountEdges',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.CountResponse.FromString,
                _registered_method=True)
        self.Stats = channel.unary_unary(
                '/swh.graph.TraversalService/Stats',
                request_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.StatsRequest.SerializeToString,
                response_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.StatsResponse.FromString,
                _registered_method=True)


class TraversalServiceServicer(object):
    """Graph traversal service 
    """

    def GetNode(self, request, context):
        """GetNode returns a single Node and its properties. 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Traverse(self, request, context):
        """Traverse performs a breadth-first graph traversal from a set of source
        nodes, then streams the nodes it encounters (if they match a given
        return filter), along with their properties.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindPathTo(self, request, context):
        """FindPathTo searches for a shortest path between a set of source nodes
        and a node that matches a specific *criteria*.

        It does so by performing a breadth-first search from the source node,
        until any node that matches the given criteria is found, then follows
        back its parents to return a shortest path from the source set to that
        node.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindPathBetween(self, request, context):
        """FindPathBetween searches for a shortest path between a set of source
        nodes and a set of destination nodes.

        It does so by performing a *bidirectional breadth-first search*, i.e.,
        two parallel breadth-first searches, one from the source set ("src-BFS")
        and one from the destination set ("dst-BFS"), until both searches find a
        common node that joins their visited sets. This node is called the
        "midpoint node".
        The path returned is the path src -> ... -> midpoint -> ... -> dst,
        which is always a shortest path between src and dst.

        The graph direction of both BFS can be configured separately. By
        default, the dst-BFS will use the graph in the opposite direction than
        the src-BFS (if direction = FORWARD, by default direction_reverse =
        BACKWARD, and vice-versa). The default behavior is thus to search for
        a shortest path between two nodes in a given direction. However, one
        can also specify FORWARD or BACKWARD for *both* the src-BFS and the
        dst-BFS. This will search for a common descendant or a common ancestor
        between the two sets, respectively. These will be the midpoints of the
        returned path.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountNodes(self, request, context):
        """CountNodes does the same as Traverse, but only returns the number of
        nodes accessed during the traversal. 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CountEdges(self, request, context):
        """CountEdges does the same as Traverse, but only returns the number of
        edges accessed during the traversal. 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Stats(self, request, context):
        """Stats returns various statistics on the overall graph. 
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TraversalServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetNode': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNode,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.GetNodeRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Node.SerializeToString,
            ),
            'Traverse': grpc.unary_stream_rpc_method_handler(
                    servicer.Traverse,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Node.SerializeToString,
            ),
            'FindPathTo': grpc.unary_unary_rpc_method_handler(
                    servicer.FindPathTo,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.FindPathToRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Path.SerializeToString,
            ),
            'FindPathBetween': grpc.unary_unary_rpc_method_handler(
                    servicer.FindPathBetween,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.FindPathBetweenRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Path.SerializeToString,
            ),
            'CountNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.CountNodes,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.CountResponse.SerializeToString,
            ),
            'CountEdges': grpc.unary_unary_rpc_method_handler(
                    servicer.CountEdges,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.CountResponse.SerializeToString,
            ),
            'Stats': grpc.unary_unary_rpc_method_handler(
                    servicer.Stats,
                    request_deserializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.StatsRequest.FromString,
                    response_serializer=swh_dot_graph_dot_grpc_dot_swhgraph__pb2.StatsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'swh.graph.TraversalService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('swh.graph.TraversalService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class TraversalService(object):
    """Graph traversal service 
    """

    @staticmethod
    def GetNode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/swh.graph.TraversalService/GetNode',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.GetNodeRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Node.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Traverse(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/swh.graph.TraversalService/Traverse',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Node.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FindPathTo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/swh.graph.TraversalService/FindPathTo',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.FindPathToRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Path.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FindPathBetween(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/swh.graph.TraversalService/FindPathBetween',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.FindPathBetweenRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.Path.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CountNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/swh.graph.TraversalService/CountNodes',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.CountResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CountEdges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/swh.graph.TraversalService/CountEdges',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.TraversalRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.CountResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Stats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/swh.graph.TraversalService/Stats',
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.StatsRequest.SerializeToString,
            swh_dot_graph_dot_grpc_dot_swhgraph__pb2.StatsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# Copyright (c) 2022-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib

from google.protobuf.field_mask_pb2 import FieldMask
import grpc
import pytest

from swh.graph.grpc.swhgraph_pb2 import GraphDirection, TraversalRequest

TEST_ORIGIN_ID = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph").hexdigest()
)
TEST_ORIGIN_ID2 = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph2").hexdigest()
)


def test_src_errors(graph_grpc_stub):
    with pytest.raises(grpc.RpcError) as exc_info:
        graph_grpc_stub.CountEdges(
            TraversalRequest(
                src=["swh:1:cnt:0000000000000000000000000000000000000404"],
                direction=GraphDirection.FORWARD,
            )
        )
    assert exc_info.value.code() == grpc.StatusCode.NOT_FOUND

    with pytest.raises(grpc.RpcError) as exc_info:
        graph_grpc_stub.CountEdges(
            TraversalRequest(
                src=["swh:1:lol:0000000000000000000000000000000000000001"],
                direction=GraphDirection.FORWARD,
            )
        )
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT

    with pytest.raises(grpc.RpcError) as exc_info:
        graph_grpc_stub.CountEdges(
            TraversalRequest(
                src=["swh:1:cnt:000000000000000000000000000000000000000z"],
                direction=GraphDirection.FORWARD,
            )
        )
    assert exc_info.value.code() == grpc.StatusCode.INVALID_ARGUMENT


def test_forward_from_root(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=[TEST_ORIGIN_ID],
            mask=FieldMask(paths=["swhid"]),
        )
    )
    assert traversal_request.count == 13


@pytest.mark.parametrize("limit", [0, 1, 2, 13, 14, 15, 1 << 63 - 1])
def test_forward_from_root_with_limit(graph_grpc_stub, limit):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=[TEST_ORIGIN_ID],
            max_matching_nodes=limit,
            mask=FieldMask(paths=["swhid"]),
        )
    )

    if limit == 1:
        assert traversal_request.count == 1
    elif limit == 2:
        assert traversal_request.count == 3
    else:
        assert traversal_request.count == 13


def test_forward_from_middle(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:dir:0000000000000000000000000000000000000012"],
            mask=FieldMask(paths=["swhid"]),
        )
    )
    assert traversal_request.count == 7


def test_forward_rel_rev(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000010"],
            edges="rel:rev,rev:rev",
            mask=FieldMask(paths=["swhid"]),
        )
    )
    assert traversal_request.count == 2


def test_backward_from_middle(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:dir:0000000000000000000000000000000000000012"],
            direction=GraphDirection.BACKWARD,
            mask=FieldMask(paths=["swhid"]),
        )
    )
    assert traversal_request.count == 6


def test_backward_from_leaf(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:cnt:0000000000000000000000000000000000000004"],
            direction=GraphDirection.BACKWARD,
            mask=FieldMask(paths=["swhid"]),
        )
    )
    assert traversal_request.count == 17


def test_backward_rev_to_rev_rev_to_rel(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:rev:0000000000000000000000000000000000000003"],
            edges="rev:rev,rev:rel",
            direction=GraphDirection.BACKWARD,
            mask=FieldMask(paths=["swhid"]),
        )
    )
    assert traversal_request.count == 6


def test_with_empty_mask(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:dir:0000000000000000000000000000000000000012"],
            mask=FieldMask(),
        )
    )
    assert traversal_request.count == 7


def test_max_depth(graph_grpc_stub):
    traversal_request = graph_grpc_stub.CountEdges(
        TraversalRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000019"],
            max_depth=1,
            mask=FieldMask(),
        )
    )
    # should count rel:19 -> rev:18
    assert traversal_request.count == 1
